"""
    Optimizer.ParamSetType.EdmParamsSheet.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy.KekLib.OurTkinter import Tk
from tksheet import Sheet
from molass_legacy._MOLASS.SerialSettings import get_setting
from .ParamsSheetBase import ParamsSheetBase

class EdmParamsSheet(ParamsSheetBase):
    def __init__(self, parent, params, dsets, optimizer):
        ParamsSheetBase.__init__(self, parent, params, dsets, optimizer)

        self.n = optimizer.n_components
        nc  = self.n - 1
        self.params_addr = [None]*len(params)
        self.params_addr_inv = {}

        body_frame = Tk.Frame(self)
        body_frame.pack()

        xr_params, xr_baseparams, rgs, mapping, uv_params, uv_baseparams, mappable_range, edm_colparams = optimizer.split_params_simple(params)

        xr_base = 0
        xr_bp_base = xr_base + np.product(xr_params.shape)
        rg_base = xr_bp_base + len(xr_baseparams)
        mp_base = rg_base + len(rgs)
        uv_base = mp_base + len(mapping)
        uv_bp_base = uv_base + len(uv_params)
        mr_base = uv_bp_base + len(uv_baseparams)
        se_base = mr_base + len(mappable_range)

        if dsets is None:
            num_columns = 8
            extra_headers = []
            xr_proportions = None
            uv_proportions = None
        else:
            from molass_legacy.Models.RateTheory.EDM import edm_impl
            num_columns = 13
            extra_headers = ["", "", "proportion"]
            a, b = mapping
            (xr_curve, xrD), rg_curve, (uv_curve, uvD) = dsets
            xr_x = xr_curve.x
            xr_proportions = []
            uv_proportions = []

            k = 0
            for t0, u, a, b, e, Dz, cinj in xr_params:
                xr_cy = edm_impl(xr_x, t0, u, a, b, e, Dz, cinj)
                xr_proportions.append(np.sum(xr_cy))
                uv_cy = uv_params[k]*xr_cy
                uv_proportions.append(np.sum(uv_cy))
                k += 1

            xr_proportions = np.array(xr_proportions)/np.sum(xr_proportions)
            uv_proportions = np.array(uv_proportions)/np.sum(uv_proportions)

        num_extended_rows = 2

        data_list = [["" for c in range(num_columns)] for r in range(self.n*2 + 2 + 8 + num_extended_rows)]

        row_offset = 0
        for j, name in enumerate(["t0", "u", "a", "b", "e", "Dz", "cinj", "",  "rg"] + extra_headers, start=1):
            data_list[row_offset][j] = name

        row_offset += 1
        n = xr_params.shape[0]
        xr_col_size = xr_params.shape[1]
        for i in range(n):
            if i == 0:
                v = "xr_params"
            else:
                v = ""
            data_list[row_offset+i][0] = v
            for j in range(xr_col_size):
                data_list[row_offset+i][j+1] = "%g" % xr_params[i,j]
                k = i*5 + j
                self.set_params_addr(xr_base+k, (row_offset+i, j+1))

            data_list[row_offset+i][xr_col_size+2] = "%g" % rgs[i]
            self.set_params_addr(rg_base+i, (row_offset+i, xr_col_size+2))

            if xr_proportions is not None:
                if i == 0:
                    data_list[row_offset+i][xr_col_size+4] = "xr area"
                if i < len(xr_proportions):
                    data_list[row_offset+i][xr_col_size+5] = "%g" % xr_proportions[i]

        row_offset += n + 1
        col_names = ["slope", "intercept"]
        if len(xr_baseparams) == 3:
            col_names.append("fouling")

        for j, name in enumerate(col_names):
            data_list[row_offset][1+j] = name

        row_offset += 1
        for i in range(1):
            if i == 0:
                v = "xr_baseline"
            else:
                v = ""
            data_list[row_offset+i][0] = v
            num_params = len(xr_baseparams)
            for j in range(num_params):
                data_list[row_offset+i][j+1] = "%g" % xr_baseparams[j]
                self.set_params_addr(xr_bp_base+j, (row_offset+i, j+1))

            # data_list[row_offset+i][xr_col_size+2] = "%g" % rgs[n+i]
            # self.set_params_addr(rg_base+n+i, (row_offset+i, xr_col_size+2))

        row_offset += 2
        data_list[row_offset][1] = "scale"
        if uv_proportions is not None:
            data_list[row_offset][12] = "proportion"

        row_offset += 1
        for i in range(n):
            if i == 0:
                v = "uv_params"
            else:
                v = ""
            data_list[row_offset+i][0] = v
            data_list[row_offset+i][1] = "%g" % uv_params[i]
            self.set_params_addr(uv_base+i, (row_offset+i, 1))

            if uv_proportions is not None:
                if i == 0:
                    data_list[row_offset+i][xr_col_size+4] = "uv area"
                if i < len(uv_proportions):
                    data_list[row_offset+i][xr_col_size+5] = "%g" % uv_proportions[i]

        row_offset += n + 1
        col_names = ["L", "x0", "k", "b", "s1", "s2", "diff_ratio"]
        if len(uv_baseparams) == 8:
            col_names.append("fouling")

        for j, name in enumerate(col_names):
            data_list[row_offset][1+j] = name

        row_offset += 1
        for i in range(1):
            if i == 0:
                v = "uv_baseline"
            else:
                v = ""
            data_list[row_offset+i][0] = v
            num_params = len(uv_baseparams)
            for j in range(num_params):
                data_list[row_offset+i][j+1] = "%g" % uv_baseparams[j]
                self.set_params_addr(uv_bp_base+j, (row_offset+i, j+1))

        row_offset += 2
        for j, name in enumerate(["slope", "intercept", "", "", "from", "to"]):
            data_list[row_offset][1+j] = name

        row_offset += 1
        data_list[row_offset][0] = "mapping"
        data_list[row_offset][4] = "mappable_range"
        for j in range(2):
            data_list[row_offset][1+j] = "%g" % mapping[j]
            self.set_params_addr(mp_base+j, (row_offset, 1+j))
            data_list[row_offset][5+j] = "%g" % mappable_range[j]
            self.set_params_addr(mr_base+j, (row_offset, 5+j))

        dark_cells = []

        if edm_colparams is not None:
            row_offset += 1
            for i, name in enumerate(["Tz"]):
                value = edm_colparams[i]
                row_offset += 1
                data_list[row_offset][0] = name
                data_list[row_offset][1] = "%g" % value
                self.set_params_addr(se_base+i, (row_offset, 1))

        self.num_valid_rows = row_offset + 2
        self.data_list = data_list
        column_width = 90
        width = column_width*num_columns + 60
        height = int(22*self.num_valid_rows) + 60
        self.sheet = Sheet(body_frame, width=width, height=height, data=data_list, show_selected_cells_border=False, column_width=column_width)

        for r, c in dark_cells:
            self.sheet.highlight_cells(r, c, bg="gray", fg="white")

        self.sheet.pack()



# coding: utf-8
"""

    ファイル名：   PilatusImage.py

    処理内容：

        シフトとマスクを考慮してピクセルの値を参照する。

    Copyright (c) 2015-2019, SAXS Team, KEK-PF
"""
from __future__ import unicode_literals
from __future__ import division
    # /   を使うと商は float
    # //  を使うと商は int
import re
from math import sqrt
import numpy as np
import pandas as pd
from OurImageIO             import Image
from SAnglerMask            import SAnglerMask
from Development            import get_devel_info
from PilatusImageProperty   import get_tiffinfo

size_of_pixel       = 172
half_size_of_pixel  = size_of_pixel // 2
comment_line_re = re.compile( '^#\s+(\w+)=(.+)' )

DEBUG = True
INVALID_VALUE = -3

def pixel_rounded_shift( delta ):
    i_ = delta // size_of_pixel
    r_ = delta %  size_of_pixel
    if r_ >= half_size_of_pixel:
        i_ += 1
    return i_

class PilatusImage:
    def __init__( self, file, mask_data='', dy=0, dx=0, i_ratio=1.0, a_ratio=0.5, original_image=None ):

        # print( 'PilatusImage: file=%s, dy=%d, dx=%d' % ( file, dy, dx ) )
        self.image      = None
        self.tiffinfo   = None

        if type(file) != np.ndarray:
            self.image      = Image( file )
            # self.tiffinfo = get_tiffinfo( im )
            self.imarray    = np.array(self.image.data)     # self.imarray.flags.writeable = True
        else:
            assert( original_image is not None )
            self.image      = original_image.image
            self.imarray    = file

        if type( mask_data ) == type( self.imarray ):
            self.mask_array = mask_data
            self.has_mask = True
        elif mask_data != '' and mask_data != None:
            self.set_mask( mask_data )
            self.has_mask = True
        else:
            self.mask_array = None
            self.has_mask = False

        if self.has_mask:
            assert self.imarray.shape == self.mask_array.shape

        self.shifted = not ( dy==0 and dx==0 )
        if self.shifted:
            self.ir = dy // size_of_pixel
            self.dr = dy %  size_of_pixel
            self.ic = dx // size_of_pixel
            self.dc = dx %  size_of_pixel
            # print 'ir=%d, dr=%d, ic=%d, dc=%d' % ( self.ir, self.dr, self.ic, self.dc )
            # pixel 単位に丸める。
            self.ir_rounded = pixel_rounded_shift( dy )
            self.ic_rounded = pixel_rounded_shift( dx )

        self.i_ratio = i_ratio

        if self.has_mask or self.shifted:
            algo = get_devel_info('adj_algorithm')
            # print( 'algo=', algo )
            if algo == 'round':
                self.round_adjust()
            elif algo == 'fast':
                self.fast_adjust( a_ratio )
            else:
                self.adjust( a_ratio )

    def image_array( self ):
        return self.imarray

    def set_mask( self, maskfile ):
        print( 'mask_array=', mask_array )
        mask_object = SAnglerMask( maskfile )
        self.mask_array = mask_object.mask_array

    def compute_cratios( self ):
        self.cratios = []
        for p, q in ( [0,0], [0,1], [1,0], [1,1] ):
            v_prop  = ( (1-p) * (size_of_pixel - self.dr) + p * self.dr ) / size_of_pixel
            h_prop  = ( (1-q) * (size_of_pixel - self.dc) + q * self.dc ) / size_of_pixel
            self.cratios.append( v_prop * h_prop )

    # --------------------------------------------------------------------------
    #   位置調整画像作成（低速版）
    # --------------------------------------------------------------------------
    def adjust( self, a_ratio ):
        if self.shifted:
            self.compute_cratios()

        imarray_ = np.zeros( self.imarray.shape )

        # TODO: マスク設定の高速化
        numrows, numcols = self.imarray.shape
        for i in range(0, numrows):
            for j in range(0, numcols):
                imarray_[i, j] = self.intensityRC( i, j, a_ratio )

        if DEBUG and self.shifted:
            # np.savetxt("slow_adjust_100.csv", imarray_[0:100,0:100], delimiter=",")
            # np.savetxt("slow_adjust_100_.csv", imarray_[0:100,-100:], delimiter=",")
            pass

        self.imarray = imarray_
        self.image.set_data( self.imarray, force=True )

    def intensityRC( self, r, c, a_ratio ):
        if not self.shifted:
            intensity = self.imarray[r,c]
            if self.has_mask and self.mask_array[ r, c ]:
                intensity = -2
        else:
            intensity = 0
            cover_ratio = 0

            k = -1
            for p, q in ( [0,0], [0,1], [1,0], [1,1] ):
                k += 1
                i = r+self.ir+p
                if not ( i >= 0 and i < self.imarray.shape[0] ):
                    continue
                j = c+self.ic+q
                if not ( j >= 0 and j < self.imarray.shape[1] ):
                    continue
                value = self.imarray[ i, j ]
                if self.has_mask and self.mask_array[ i, j ]:
                    value = -2

                if value < 0:
                    continue

                ratio = self.cratios[ k ]
                intensity += value * ratio
                cover_ratio += ratio

            if cover_ratio < a_ratio:
                intensity = -3
            elif cover_ratio < 1.0:
                    # print 'cover_ratio(%d,%d)=%g' % ( r, c, cover_ratio )
                    intensity /= cover_ratio

        if intensity > 0:
            intensity /= self.i_ratio

        return int( intensity )

    # --------------------------------------------------------------------------
    #   位置調整画像作成（移動量 pixel 丸め版、最高速）
    # --------------------------------------------------------------------------
    def round_adjust( self ):
        # マスクをかける。
        if self.has_mask:
            self.imarray[ self.mask_array == 1 ] = -2

        if not self.shifted:
            return

        # print( 'ir_=%d, ic_=%d' % ( self.ir_rounded, ic_ ) )
        # 対応不能領域を含めて全体に無効値（-2）を設定しておく。
        adjusted_array = np.ones( self.imarray.shape ) * ( -2 )

        # 対応可能な矩形領域の両端を把握する。
        row, col = self.imarray.shape
        adj_min_r = max(   0,   0 - self.ir_rounded )
        adj_max_r = min( row, row - self.ir_rounded )
        adj_min_c = max(   0,   0 - self.ic_rounded )
        adj_max_c = min( col, col - self.ic_rounded )
        sft_min_r = max(   0,   0 + self.ir_rounded )
        sft_max_r = min( row, row + self.ir_rounded )
        sft_min_c = max(   0,   0 + self.ic_rounded )
        sft_max_c = min( col, col + self.ic_rounded )

        # 行列（numpy.ndarray）のスライスを使ってデータを対応位置にコピーする。
        adjusted_array[ adj_min_r:adj_max_r, adj_min_c:adj_max_c ] = self.imarray[ sft_min_r:sft_max_r, sft_min_c:sft_max_c ]
        self.imarray = adjusted_array.astype( 'i4' )
        self.image.set_data( self.imarray, force=True )

        return

    # --------------------------------------------------------------------------
    #   位置調整画像作成（高速版）
    # --------------------------------------------------------------------------
    def fast_adjust( self, a_ratio ):
        if self.has_mask:
            self.imarray[ self.mask_array == 1 ] = -2

        if not self.shifted:
            return

        # print 'Doing fast_adjust'

        # ４つの成分行列 (0,0), (0,1), (1,0), (1,1) のカバー比率を計算しておく。
        self.compute_cratios()

        # 周囲の境界部分を上下左右に 1 pixel 幅だけ拡張し、無効値を設定した拡張画像データを作る。
        numrows, numcols = self.imarray.shape
        extended_array = np.ones( [ numrows+2, numcols+2 ] ) * INVALID_VALUE
        extended_array[1:numrows+1,1:numcols+1] = self.imarray

        # シフト量を考慮して、成分行列 (0,0) 相当の参照可能範囲を把握する。
        a00_r_min = max(       0 + self.ir,           0 )
        a00_r_max = min( numrows + self.ir + 1, numrows + 1 )
        a00_c_min = max(       0 + self.ic,           0 )
        a00_c_max = min( numcols + self.ic + 1, numcols + 1 )

        # ４つの成分行列 (0,0), (0,1), (1,0), (1,1) から値行列とウェイト行列を計算する。
        k = 0
        for p, q in ( [0,0], [0,1], [1,0], [1,1] ):
            carray_ = extended_array[ ( a00_r_min+p ):( a00_r_max+p ), ( a00_c_min+q ):( a00_c_max+q ) ] * ( self.cratios[ k ] / self.i_ratio )
            # print 'carray[%d].shape=' % ( k ), carray_.shape

            if k == 0:
                v_array  = np.zeros( carray_.shape, dtype='f8' )    # 作業用の調整値行列
                w_array  = np.zeros( carray_.shape, dtype='f8' )    # ウェイト行列

            valid_cond_ = carray_ >= 0
            v_array[ valid_cond_ ] += carray_[ valid_cond_ ]
            w_array[ valid_cond_ ] += self.cratios[ k ]

            k += 1

        # ウェイトが許容可能なカバー比率未満の pixel の値を無効値にする。
        v_array[ w_array < a_ratio ] = INVALID_VALUE

        # ウェイトが許容可能なカバー比率以上の pixel の値を基準化する。
        valid_cond = w_array >= a_ratio
        v_array[ valid_cond ] /= w_array[ valid_cond ]

        # 結果の調整行列における設定位置を把握する。
        # TODO: check for both signs (+-) of shift values
        b00_r_min = max(       0 - self.ir,       0 )
        if b00_r_min > 0:
            b00_r_min   -= 1
            r_offset = 0
        else:
            r_offset = 1
        b00_r_max = b00_r_min + v_array.shape[0] - r_offset
        b00_c_min = max(       0 - self.ic,       0 )
        if b00_c_min > 0:
            b00_c_min   -= 1
            c_offset = 0
        else:
            c_offset = 1
        b00_c_max = b00_c_min + v_array.shape[1] - c_offset

        # print b00_r_min, b00_r_max, b00_c_min, b00_c_max

        # 結果の調整行列を無効値（int型）で初期化する。
        imarray_ = np.ones( self.imarray.shape, dtype='i4' ) * INVALID_VALUE

        # 上で計算した調整値行列を結果の調整行列の該当位置に代入する。
        imarray_[ b00_r_min:b00_r_max, b00_c_min:b00_c_max ] = v_array[r_offset:v_array.shape[0], c_offset:v_array.shape[1]]

        if DEBUG:
            # np.savetxt("fast_adjust.csv", imarray_, delimiter=",")
            # np.savetxt("fast_adjust_100.csv", imarray_[0:100,0:100], delimiter=",")
            # np.savetxt("fast_adjust_100_.csv", imarray_[0:100,-100:], delimiter=",")
            pass

        self.imarray = imarray_     # dtype='i4' のままでよい。
        self.image.set_data( self.imarray, force=True )

    # --------------------------------------------------------------------------
    #   ２画像の欠損補填（廃棄予定）
    # --------------------------------------------------------------------------
    def make_covered_array( self, other ):

        assert self.imarray.shape == other.imarray.shape
        print( 'Making covered_array' )

        covered_array = self.imarray.copy()

        numrows, numcols = self.imarray.shape
        for i in range(0, numrows):
            for j in range(0, numcols):
                intensity0 = self.imarray[ i, j ]
                intensity1 = other.imarray[ i, j ]
                if intensity0 < 0:
                    if intensity1 >= 0:
                        # other 側に有効値があればそれを設定する。
                        covered_array[ i, j ] = intensity1
                    else:
                        # other 側に有効値がなければ、自身のマスク示唆値（負値）を設定する。
                        covered_array[ i, j ] = intensity0

        return covered_array.astype('i4')

    # --------------------------------------------------------------------------
    #   ２画像の欠損補填
    # --------------------------------------------------------------------------
    def fast_make_covered_array( self, other ):

        assert self.imarray.shape == other.imarray.shape
        # print 'Fast making covered_array'

        covered_array = self.imarray.copy()

        valid_cond = np.logical_and( self.imarray < 0, other.imarray >= 0 )

        covered_array[ valid_cond ] = other.imarray[ valid_cond ]

        return covered_array.astype('i4')

    # --------------------------------------------------------------------------
    #   ２画像の平均化（廃棄予定）
    # --------------------------------------------------------------------------
    def make_average_array( self, other ):

        assert self.imarray.shape == other.imarray.shape
        print( 'Making average_array' )

        average_array = self.imarray.copy()

        numrows, numcols = self.imarray.shape
        for i in range(0, numrows):
            for j in range(0, numcols):
                intensity0 = self.imarray[ i, j ]
                intensity1 = other.imarray[ i, j ]
                if intensity0 < 0:
                    if intensity1 >= 0:
                        # other 側に有効値があればそれを設定する。
                        average_array[ i, j ] = intensity1
                    else:
                        # other 側に有効値がなければ、自身のマスク示唆値（負値）を設定する。
                        average_array[ i, j ] = intensity0
                else:
                    if intensity1 >= 0:
                        average_array[ i, j ] = ( intensity0 + intensity1 ) // 2
                    else:
                        average_array[ i, j ] = intensity0

        return average_array.astype('i4')

    # --------------------------------------------------------------------------
    #   多画像の平均化
    # --------------------------------------------------------------------------
    def fast_make_average_array( self, counter_array, sum_array ):

        # 無効値で初期化する。
        im_array  = np.ones( counter_array.shape, dtype='i4' ) * INVALID_VALUE

        # 有効値が少なくとも１つあった pixel について平均を計算する。
        valid_cond = counter_array > 0
        im_array[ valid_cond ] = sum_array[ valid_cond ] / counter_array[ valid_cond ]

        return im_array

    # --------------------------------------------------------------------------
    #   テスト支援メソッド群
    # --------------------------------------------------------------------------
    def diff( self, other ):
        diff_imarray = np.zeros( self.imarray.shape )
        diff_imarray[ self.imarray != other.imarray ] = 10000
        return PilatusImage( diff_imarray, original_image=self )

    def save( self, path ):
        assert( self.image is not None )
        self.image.save( path )

    def equal( self, other ):
        diff_im_array = self.diff( other ).image_array()
        isequal_ = ( diff_im_array == np.zeros( self.imarray.shape ) ).all()
        if not isequal_:
            print( diff_im_array == np.zeros( self.imarray.shape ) )
            try:
                for i in range( diff_im_array.shape[0] ):
                    for j in range( diff_im_array.shape[1] ):
                        if diff_im_array[ i, j ] == 10000:
                            print( 'self[%d,%d]=%d, other[%d,%d]=%d' % ( i, j, self.imarray[i,j], i, j, other.imarray[i,j]) )
                            assert( False )
            except:
                pass
        return isequal_



# coding: utf-8
"""
    ExSolver.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
from SvdDenoise import get_denoised_data

class ExSolver:
    def __init__(self):
        pass

    def solve(self, M, c, cd=1):
        M_ = get_denoised_data(M, rank=cd)
        if cd == 1:
            C = np.array([c])
        elif cd == 2:
            C = np.array([c, c**2])
        else:
            assert False

        P = M_ @ np.linalg.pinv(C)
        return P[:,0]



"""

    Version.py

    Copyright (c) 2016-2025, SAXS Team, KEK-PF

"""
import platform
from molass_legacy import get_version

def get_version_string(cpuid=False):
    if cpuid:
        from molass_legacy.KekLib.MachineTypes import get_cpuid
        cpuid = ' cpuid:' + str(get_cpuid())
    else:
        cpuid = ''

    version = get_version()
    return 'MOLASS %s (python %s %s%s)' % (
                version, platform.python_version(), platform.architecture()[0], cpuid )

def molass_version_for_publication():
    import re
    version = get_version_string()
    return re.sub(r"\s+\(.+", "", version)

def is_developing_version():
    return get_version_string().find(":") > 0



# coding: utf-8
"""
    GaussianMixture.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np

# gaussian pdf function
SQRT_2PI = np.sqrt(2*np.pi)
def gaussian_pdf(x, m, s, scale=1):
    return scale * 1/(s*SQRT_2PI) * np.exp(-0.5*((x-m)/s)**2)

# mixture curve data
PARAMS = [(100, 30, 1), (200, 20, 0.5)]

def gm_curve(x_length = 300, params=PARAMS, ret_detail=False):
    x = np.arange(x_length)
    gy = [gaussian_pdf(x,*p) for p in params]
    y = np.sum(gy, axis=0)
    if ret_detail:
        return x, y, gy
    else:
        return x, y

def hist_to_source(x, y, num_presision = 100):
    # conversion from curve (histogram) to source data
    data = []
    max_y = np.max(y)
    for x_, y_ in zip(x, y):
        data += [x_]*int(y_/max_y*num_presision)

    sy = np.array(data)
    return sy

def get_sorted_params(gmm):
    # sort results in the order of the means
    params = zip(gmm.means_.ravel(), gmm.covariances_.ravel(), gmm.weights_)
    sorted_params = sorted([ (m, v, w) for m, v, w in params], key=lambda p:p[0])
    return sorted_params

def get_curves(gmm, x):
    sorted_params = get_sorted_params(gmm)

    # plot the results
    gy_list = []
    ty = np.zeros(len(x))
    for m, v, w in sorted_params:
        gy = gaussian_pdf(x, m, np.sqrt(v), w)
        gy_list.append(gy)
        ty += gy

    return ty, gy_list

def debug_plot_gmm(x, y, sy, gmm, gmm2=None, num_presision = 100):
    import molass_legacy.KekLib.DebugPlot as plt
    max_y = np.max(y)

    if gmm2 is None:
        fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(12,5))
        ax1, ax3 = axes
        ax2 = None
        pmm_plot_params = [(ax3, gmm)]
    else:
        fig, axes = plt.subplots(nrows=1, ncols=3, figsize=(18,5))
        ax1, ax2, ax3 = axes
        pmm_plot_params = [(ax2, gmm2), (ax3, gmm)]

    ax1.set_title("Source Data Generation")
    if ax2 is not None:
        ax3.set_title("Gaussian Mixture Model Fitting (intermediate)")
    ax3.set_title("Gaussian Mixture Model Fitting")

    ax1t = ax1.twinx()
    ax1.plot(y, linewidth=3, label='given curve')
    ax1t.hist(sy, bins=len(x), color='orange', label='histogram of generated data')

    fig_scale = 1.05
    ax1.set_ylim(0, max_y*fig_scale)
    ax1t.set_ylim(0, num_presision*fig_scale)

    ax1.legend(bbox_to_anchor=(1, 0.98), loc='upper right')
    ax1t.legend(bbox_to_anchor=(1, 0.92), loc='upper right')

    for ax, gmm_ in pmm_plot_params:
        ty, gy_list = get_curves(gmm_, x)
        ax.plot(y, linewidth=3, label='given curve')
        axt = ax.twinx()
        axt.plot(ty, ':', color='red', linewidth=3, label='GMM fitted curve')
        for k, gy in enumerate(gy_list):
            axt.plot(gy, ':', label='%d-th component' % k)

        ax.legend(bbox_to_anchor=(1, 0.98), loc='upper right')
        axt.legend(bbox_to_anchor=(1, 0.92), loc='upper right')

    fig.tight_layout()
    plt.show()



"""
    SimpleGuinierDemo.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import os
from bisect import bisect_right
import numpy as np
import time
import molass_legacy.KekLib.DebugPlot as plt
from DataUtils import get_in_folder
from SvdDenoise import get_denoised_data
from SimpleGuinier import SimpleGuinier

def compute_lrf_info(D, E, qv, i_smp, aslice, eslice, rank=2):
    D_ = D[:,eslice]
    E_ = E[:,eslice]

    D_ = get_denoised_data(D_, rank=rank)
    c = D_[i_smp, :]
    if rank == 1:
        C = np.array([c])
    else:
        C = np.array([c, c**2])
    Cinv = np.linalg.pinv(C)
    P = D_ @ Cinv
    Dinv = np.linalg.pinv(D_)
    W = np.dot(Dinv, P)
    Pe = np.sqrt(np.dot(E_**2, W**2))
    data = np.array([qv, P[:,0], Pe[:,0]]).T
    sg = SimpleGuinier(data)
    z_ = np.log(data[aslice,1])
    return sg, z_

g_counter = 0

def demo(in_folder, sd, fig_folder=None, counter=None, for_all_peaks=False, index_csv_fh=None):

    print(in_folder)

    if counter is None:
        global g_counter
        counter = g_counter
        g_counter += 1

    if index_csv_fh is not None:
        index_csv_fh.write(','.join([str(counter), in_folder]) + "\n")
        index_csv_fh.flush()

    D, E, qv, ecurve = sd.get_xr_data_separate_ly()

    for k, pinfo in enumerate(ecurve.peak_info):
        if for_all_peaks:
            pass
        else:
            if k != ecurve.primary_peak_no:
                continue

        eno = pinfo[1]
        eslice = slice(pinfo[0], pinfo[2]+1)

        y = D[:,eno]
        e = E[:,eno]
        data = np.array([qv, y, e]).T
        sg = SimpleGuinier(data)
        i_smp = sd.xray_index

        i = bisect_right(qv, 2/max(10, sg.Rg))

        aslice = slice(0, i)
        qv2 = qv[aslice]**2
        y_ = np.log(y[aslice])

        lrf1_sg, z1_ = compute_lrf_info(D, E, qv, i_smp, aslice, eslice, rank=1)

        lrf2_sg, z2_ = compute_lrf_info(D, E, qv, i_smp, aslice, eslice, rank=2)

        peak_suffix = "-%d" % k if for_all_peaks else ""

        with plt.Dp():
            in_folder_ = get_in_folder(in_folder)
            fig, ax = plt.subplots()

            ax.set_title("Guinier Plots of Curves in %s%s" % (in_folder_, peak_suffix))

            Iz = np.log(sg.Iz)
            ax.plot(qv2, y_ - Iz, "o", markersize=1, label="main-peak curve")
            ax.plot(sg.guinier_x, sg.guinier_y - Iz, color="red", alpha=0.5, lw=2, label="main-peak guinier region")

            for lrf_sg, z_, rank in [(lrf1_sg, z1_, 1), (lrf2_sg, z2_, 2)]:
                Iz = np.log(lrf_sg.Iz)
                ax.plot(qv2, z_ - Iz, "o", markersize=1, label="rank%d-lrf curve" % rank)
                color = "pink" if rank == 1 else "cyan"
                ax.plot(lrf_sg.guinier_x, lrf_sg.guinier_y - Iz, color=color, alpha=0.5, lw=2, label="rank%d-lrf guinier region" % rank)

            ax.legend()

            xmin, xmax = ax.get_xlim()
            ymin, ymax = ax.get_ylim()
            ax.set_ylim(min(-1.2, ymin), ymax)
            tx = xmin*0.9 + xmax*0.1
            w = 0.3; ty = ymin*(1-w) + ymax*w
            ax.text(tx, ty, "Peak Rg=%.1f min_qRg=%.3g" % (sg.Rg, sg.min_qRg), fontsize=16, alpha=0.5, va="center")
            w = 0.2; ty = ymin*(1-w) + ymax*w
            ax.text(tx, ty, "LRF1 Rg=%.1f min_qRg=%.3g" % (lrf1_sg.Rg, lrf1_sg.min_qRg), fontsize=16, alpha=0.5, va="center")
            w = 0.1; ty = ymin*(1-w) + ymax*w
            ax.text(tx, ty, "LRF2 Rg=%.1f min_qRg=%.3g" % (lrf2_sg.Rg, lrf2_sg.min_qRg), fontsize=16, alpha=0.5, va="center")

            fig.tight_layout()

            if fig_folder is None:
                plt.show()
            else:
                plt.show(block=False)
                dp = plt.get_dp()
                fig_file = os.path.join(fig_folder, "fig-%03d%s.png" % (counter, peak_suffix))
                dp.fig.savefig(fig_file)
                time.sleep(1)



"""
    Optimizer.OptDataSets.py

    Copyright (c) 2022-2024, SAXS Team, KEK-PF
"""
import os
import logging
import numpy as np
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting
from .TheUtils import get_optimizer_folder
from molass_legacy.RgProcess.RgCurve import check_rg_folder, RgCurve
from molass_legacy.RgProcess.RgCurveProxy import RgCurveProxy
import molass_legacy.KekLib.DebugPlot as plt

def get_current_rg_folder(compute_rg=False, possibly_relocated=True, current_folder=None):
    """
    arguments will be used as follows.

    process                  compute_rg  possibly_relocated  current_folder
    ----------------------------------------------------------------------------
    PeakEditor.                 True            False             None
    OptimizerUtils              False           False             None
    FullOptDialog
        (recompute rg-curve)    True            True              specified
    ResultFolderSelector                False           True              specified
    OptimizerMain               False           True              None (getcwd())

    """
    optimizer_folder = get_optimizer_folder()
    rg_folder = os.path.join(optimizer_folder, 'rg-curve')
    if os.path.exists(rg_folder):
        return rg_folder

    if possibly_relocated:
        pass
    else:
        if compute_rg:
            return rg_folder

    relocated_root_folder = rg_folder + "s"
    assert os.path.exists(relocated_root_folder)

    if current_folder is None:
        current_folder = os.getcwd()

    _, jobno = os.path.split(current_folder)
    relocated_folder = os.path.join(relocated_root_folder, jobno)
    if compute_rg:
        pass
    else:
        assert os.path.exists(relocated_folder)

    return relocated_folder

def get_dsets_impl(sd, corrected_sd, progress_cb=None, rg_folder=None, rg_info=True, logger=None,
                    compute_rg=False, possibly_relocated=True, current_folder=None):
    if logger is not None:
        logger = logging.getLogger(__name__)

    D, E, qv, xr_curve = sd.get_xr_data_separate_ly()
    xr_curve_ = None

    if rg_info:
        if rg_folder is None:
            rg_folder = get_current_rg_folder(compute_rg=compute_rg, possibly_relocated=possibly_relocated, current_folder=current_folder)

        rg_folder_ok = check_rg_folder(rg_folder)
        logger.info("rg_folder=%s status is %s", rg_folder, str(rg_folder_ok))

        if not rg_folder_ok:
            previous_rg_folder = get_setting("rg_curve_folder")
            if previous_rg_folder is None:
                # i.e., not ok
                pass
            else:
                if check_rg_folder(previous_rg_folder):
                    from shutil import rmtree, copytree
                    if os.path.exists(rg_folder):
                        # maybe broken
                        rmtree(rg_folder)
                    copytree(previous_rg_folder, rg_folder)
                    rg_folder_ok = check_rg_folder(rg_folder)
                    logger.info("rg-curve has been copied from %s: check status %s", previous_rg_folder, str(rg_folder_ok))

        if rg_folder_ok:
            rg_curve = RgCurveProxy(xr_curve, rg_folder, progress_cb=progress_cb)
            trust_rg_curve_folder = get_setting("trust_rg_curve_folder")
            if trust_rg_curve_folder:
                logger.info("rg-curve proxy is trusted without trimming consistency check.")
            else:
                xr_restrict_list = get_setting("xr_restrict_list")
                current_rg_trimming = str(None if xr_restrict_list is None else xr_restrict_list[0])
                proxy_rg_trimming = str(rg_curve.rg_trimming)
                logger.info("trimmings are %s, %s", current_rg_trimming, proxy_rg_trimming)
                if current_rg_trimming == proxy_rg_trimming:
                    logger.info("rg-curve has been created as a proxy.")
                else:
                    logger.info("the existing rg-curve will be discarded due to trimming inconsistency.")
                    rg_folder_ok = False

        if not rg_folder_ok:
            # create rg_curve from corrected_sd
            D_, E_, qv_, xr_curve_ = corrected_sd.get_xr_data_separate_ly()
            rg_curve = RgCurve(qv_, xr_curve_, D_, E_, progress_cb=progress_cb)

            from molass_legacy.KekLib.BasicUtils import clear_dirs_with_retry
            clear_dirs_with_retry([rg_folder])
            rg_curve.export(rg_folder)
            set_setting("rg_curve_folder", rg_folder)
            logger.info("rg-curve calculation complete.")
    else:
        rg_curve = None

    U, _, wlvec, uv_curve = sd.get_uv_data_separate_ly()

    if False:
        with plt.Dp():
            fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
            fig.suptitle("get_dsets_impl: final")
            ax1.plot(uv_curve.x, uv_curve.y)
            ax2.plot(xr_curve.x, xr_curve.y)
            if xr_curve_ is not None:
                ax2.plot(xr_curve_.x, xr_curve_.y)
            fig.tight_layout()
            plt.show()

    return ((xr_curve, D), rg_curve, (uv_curve, U))

class OptDataSets:
    def __init__(self, sd, corrected_sd, dsets=None, rg_folder=None, progress_cb=None, compute_rg=False, possibly_relocated=True, current_folder=None):
        self.logger = logging.getLogger(__name__)
        if dsets is None:
            dsets = get_dsets_impl(sd, corrected_sd, progress_cb=progress_cb, rg_folder=rg_folder, logger=self.logger,
                                    compute_rg=compute_rg, possibly_relocated=possibly_relocated, current_folder=current_folder)
        self.dsets = dsets
        try:
            D = dsets[0][1]
            E = sd.intensity_array[:,:,2].T
            self.weight_info = self.compute_weight_info(1/(E + D/100))
        except:
            # LinAlgError("SVD did not converge") as in 20200214_2
            from molass_legacy.KekLib.ExceptionTracebacker import log_exception
            log_exception(self.logger, "compute_weight_info error: ")

            D, E, qv, curve = sd.get_xr_data_separate_ly()
            self.weight_info = self.compute_weight_info(1/(E + D/100))  # seems to work for 20200214_2
            self.logger.info("retried with self.compute_weight_info(1/(E + np.abs(D)/100))")

            if False:
                # this plot failes to appear
                import molass_legacy.KekLib.DebugPlot as plt
                from MatrixData import simple_plot_3d
                with plt.Dp():
                    fig, (ax1, ax2)= plt.subplots(ncols=2, figsize=(12,5), subplot_kw=dict(projection="3d"))
                    fig.suptitle("compute_weight_info Error")
                    simple_plot_3d(ax1, D, x=qv)
                    simple_plot_3d(ax2, E, x=qv)
                    fig.tight_layout()
                    plt.show()
        self.E = E

    def relocate_rg_folder(self):
        recompute_rg_curve = get_setting("recompute_rg_curve")
        if recompute_rg_curve:
            import shutil
            from molass_legacy.Optimizer.TheUtils import get_optimizer_folder
            from molass_legacy.KekLib.BasicUtils import clear_dirs_with_retry

            optimizer_folder = get_optimizer_folder()
            rg_folder = os.path.join(optimizer_folder, "rg-curve")
            assert os.path.exists(rg_folder)

            rg_curves_folder = os.path.join(optimizer_folder, "rg-curves")
            clear_dirs_with_retry([rg_curves_folder])
            shutil.move(rg_folder, rg_curves_folder)

            src_path = os.path.join(rg_curves_folder, "rg-curve")
            tgt_path = os.path.join(rg_curves_folder, "000")
            os.rename(src_path, tgt_path)
            self.logger.info("rg-curve has been relocated to %s", tgt_path)

    def compute_weight_info(self, W):
        Wn = W/np.std(W)
        U, s, VT = np.linalg.svd(Wn)
        rank = 1
        s_ = np.sqrt(s[0])
        uw = U[:,0:rank]*s_
        vw = VT[0:rank,:]*s_
        W_ = uw @ vw
        return uw, vw, W_

    def __iter__(self):
        # this is for backward compatibility
        return iter(self.dsets)

    def __getitem__(self, item):
        # this is for backward compatibility
        return self.dsets[item]

    def get_opt_weight_info(self):
        return self.weight_info



"""
    CharFunc/CfDomainOptimizer.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import minimize

class CfDomainOptimizer:
    def __init__(self, x, y, debug=False):
        pass

    


"""
    Optimizer.ScoreTransition.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.Optimizer.NaviFrame import NaviFrame
from molass_legacy.Optimizer.FvScoreConverter import convert_score

class ScoreTransition(Dialog):
    def __init__(self, parent, js_canvas, optimizer, x_array, fv_array, best_index):
        self.optimizer = optimizer
        self.js_canvas = js_canvas
        self.x_array = x_array
        self.fv_scores = fv_array[:,1]
        self.best_index = best_index
        Dialog.__init__(self, parent, "Score Transition", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        cframe = Tk.Frame(body_frame)
        cframe.pack()
        bframe = Tk.Frame(body_frame)
        bframe.pack()

        gs = GridSpec(8,1)

        fig = plt.figure(figsize=(8,8))
        ax0 = fig.add_subplot(gs[0:7,:])
        ax1 = fig.add_subplot(gs[7,:])
        # ax1.set_axis_off()
        self.fig = fig
        self.axes = ax0, ax1
        self.mpl_canvas = FigureCanvasTkAgg(fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1)
        self.draw_score(self.best_index)

        nframe = NaviFrame(bframe, self, arrows_only=True)
        nframe.pack()

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack()

        w = Tk.Button(box, text="Close", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=50, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def get_params(self, index):
        return self.x_array[index]

    def draw_score(self, index, devel=True):
        if devel:
            import Optimizer.TheDebugUtils
            from importlib import reload
            reload(Optimizer.TheDebugUtils)
        from molass_legacy.Optimizer.TheDebugUtils import plot_scores

        self.curr_index = index

        fig = self.fig
        ax0, ax1 = self.axes
        optimizer = self.optimizer
        # in_folder, seeds, result = fig_info

        score_names = optimizer.get_score_names()

        params = self.get_params(self.best_index)
        fv, score_list = optimizer.objective_func(params, return_full=True)[0:2]
        ax0.cla()
        ax1.cla()
        ax1.set_ylim(0, 1)
        ax1.yaxis.tick_right()
        ax1.yaxis.set_label_position("right")
        ax1.set_yticks([0.5], ["evaluation position"])

        i = self.best_index
        ax1.plot([i,i], [0,1], color="red", lw=2, label="best [%d]" % self.best_index)

        dx = len(self.fv_scores)*0.05
        ax1.set_xlim(-dx, len(self.fv_scores)+dx)

        if index == self.best_index:
            score = convert_score(self.fv_scores[index])
            ax0.set_title("Score at Best %.1f at [%d]" % (score, index), fontsize=16)
        else:
            if index < self.best_index:
                i, j = (index, self.best_index)
            else:
                i, j = (self.best_index, index)
            scores = [convert_score(self.fv_scores[k]) for k in [i, j]]
            ax0.set_title("Score Transition from %.1f at [%d] to %.1f at [%d]" % (scores[0], i, scores[1], j), fontsize=16)
            ax1.plot([index,index], [0,1], color="yellow", lw=2, label="current [%d]" % index)

        if index == self.best_index:
            plot_scores(ax0, score_list, score_names, label="best [%d]" % self.best_index)
        else:
            plot_scores(ax0, score_list, score_names, label="best [%d]" % self.best_index, alpha=0.5)
            params = self.get_params(index)
            fv, score_list = optimizer.objective_func(params, return_full=True)[0:2]
            plot_scores(ax0, score_list, score_names, label="current [%d]" % self.curr_index, alpha=0.5, invert=False, add_patch=False)
        ax0.legend(loc="lower right")
        ax1.legend()

        fig.tight_layout()
        self.mpl_canvas.draw()

    def get_first(self):
        self.draw_score(0)

    def get_previous_best(self):
        index = self.get_best_index(stop=self.curr_index)
        self.draw_score(index)

    def get_previous(self):
        if self.curr_index > 0:
            index = self.curr_index - 1
        else:
            index = self.curr_index
        self.draw_score(index)

    def get_best_index(self, stop=None):
        m = np.argmin(self.fv_scores[:stop,1])
        return m

    def get_best(self):
        self.draw_score(self.best_index)

    def get_next(self):
        if self.curr_index < len(self.fv_scores) - 1:
            index = self.curr_index + 1
        else:
            index = self.curr_index
        self.draw_score(index)

    def get_next_best(self):
        fv_scores = self.fv_scores
        stop = min(len(fv_scores), self.curr_index + 1)
        best_fv_until_now = np.min(fv_scores[:stop,1])
        w  = np.where(fv_scores[stop:,1] < best_fv_until_now)[0]
        if len(w) > 0:
            self.draw_score(stop + w[0])

    def get_last(self):
        self.draw_score(len(self.fv_scores)-1)



# coding: utf-8
"""

    OpenPyXlUtil.py

    Copyright (c) 2016-2017, Masatsuyo Takahashi, KEK-PF

"""

"""
from https://msdn.microsoft.com/en-us/library/bb264112(v=vs.85).aspx

EMU - english metrical unit
There are 914400 EMU per inch and 12700 EMU in a point.
"""
LINE_WIDTH      = int( 12700 * 1.5 )

def save_allowing_user_reply( wb, xlsx_file, parent=None ):
    ok = False

    while not ok:
        try:
            wb.save( xlsx_file )
        except Exception as ex:
            if str( ex ).find( 'Permission' ) >= 0:
                import molass_legacy.KekLib.OurMessageBox as MessageBox
                oc = MessageBox.askokcancel(
                        'Permission Error Retry',
                        "'%s' is being used by another application.\nClose it and press 'OK' if you wish retry." % xlsx_file,
                        parent=parent )
                if oc:
                    continue
        ok = True



"""
    SecTheory.Edm.py

    EDM - Equilibrium Dispersive Model

    based on the paper 2022, Jamil Ur Rehman,
        ANALYSIS OF EQUILIBRIUM DISPERSIVE MODEL
        OF LIQUID CHROMATOGRAPHY CONSIDERING
        A QUADRATIC-TYPE ADSORPTION ISOTHERM

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.special import erfc
from scipy.optimize import minimize, basinhopping

class Edm:
    def __init__(self, z=30, L=30, a=1.5, b=0.5, u=1.2, e=0.4, Dz=0.02, cinit=0, cinj=1.0, c0=0.0001, tinj=2.0):
        self.a = a
        self.b = b
        self.e = e
        self.Dz = Dz
        self.cinj = cinj
        self.F = F = (1 - e)/e      # phase ratio
        self.L = L
        self.x = z/L
        self.u = u
        self.Pe = Pe = L*u/Dz   # 

        denominator = (1 + b*c0)**3
        gam1 = (a * b**2 * c0**3) / denominator
        gam2 = (a + 3*a*b*c0) / denominator
        gam3 = (-a*b) / denominator
        self.R = R = 1 + gam2*F
        self.lam = lam = 2*F*gam3/R
        tau_inj = u*tinj/L
        self.Beta = lam*Pe/2 * cinj * tau_inj
        self.RPe = R*Pe

    def get_comp_params(self):
        return self.a, self.b, self.e, self.Dz, self.cinj

    def __call__(self, t):
        x = self.x
        L = self.L
        u = self.u
        tau = u*t/L
        xi = x - tau/self.R
        Pe = self.Pe

        U = 2/(self.lam*Pe)
        RPe = self.RPe
        V = xi**2/(4*tau/RPe)
        W = np.sqrt(np.pi*tau/RPe)
        Y = xi/(2*np.sqrt(tau/RPe))

        numerator = ( -U*np.exp(-V) + U*np.exp(self.Beta - V) )
        denominator = ( W*np.exp(self.Beta)* erfc(-Y) + W*erfc(Y) )

        ret_y = numerator / denominator
        ret_y[np.isnan(ret_y)] = 0
        return ret_y

NEGATIVE_PENALTY_SCALE = 1000

def guess_single_edm(x, y, init_params=None, debug=False):

    z = 30
    L = 30
    u = 0.5

    def objective(p):
        a, b, e, Dz, cinj = p
        edm = Edm(z=z, L=L, a=a, b=b, u=u, e=e, Dz=Dz, cinj=cinj)
        y_ = edm(x)
        negative = NEGATIVE_PENALTY_SCALE*min(0, Dz)**2
        fv = np.sum((y_ - y)**2) + negative
        return fv

    if init_params is None:
        init_params = (1.5, -3.0, 0.4, 0.06, 0.5)

    ret = minimize(objective, init_params)
    a, b, e, Dz,cinj = ret.x
    print("params=", (a, b, e, Dz, cinj))
    # [1]  (1.5047442427375226, -2.627270376342513,  0.3861655349460303,  0.05006483813658697,  0.584771571520628)
    # [2]  (1.4479395094650214, -3.0295291424530553, 0.25816194967491296, 0.011396843248255373, 0.052114693543924236)
    edm = Edm(z=z, L=L, a=a, b=b, u=u, e=e, Dz=Dz, cinj=cinj)

    if debug:
        import molass_legacy.KekLib.DebugPlot as plt
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.plot(x, y, label="data")
            ax.plot(x, edm(x), label="model")
            ax.legend()
            fig.tight_layout()
            plt.show()

    return edm

def guess_multiple_edms(x, y, num_components, debug=False):
    models = []
    y_ = y.copy()
    for k in range(num_components):
        edm = guess_single_edm(x, y_, debug=debug)
        models.append(edm)
        if len(models) == num_components:
            break

        cy = edm(x)
        y_ -= cy
    return models

def edm_func(t, u, a, b, e, Dz, cinj, cinit=0, c0=0.0001, tinj=2.0, L=30, z=30):
    F = (1 - e)/e      # phase ratio
    x = z/L
    Pe = L*u/Dz   # 

    denominator = (1 + b*c0)**3
    gam2 = (a + 3*a*b*c0) / denominator
    gam3 = (-a*b) / denominator
    R = 1 + gam2*F
    lam = 2*F*gam3/R
    tau_inj = u*tinj/L
    Beta = lam*Pe/2 * cinj * tau_inj
    RPe = R*Pe

    tau = u*t/L
    xi = x - tau/R

    U = 2/(lam*Pe)
    V = xi**2/(4*tau/RPe)
    W = np.sqrt(np.pi*tau/RPe)
    Y = xi/(2*np.sqrt(tau/RPe))

    # numer = U*( -np.exp(-V) + np.exp(Beta - V) )
    # denom = W*( np.exp(Beta)* erfc(-Y) + erfc(Y) )
    expB = np.exp(Beta)
    numer = U*(expB - 1)*np.exp(-V)
    denom = W*( expB * erfc(-Y) + erfc(Y) )

    ret_y = numer / denom
    ret_y[np.isnan(ret_y)] = 0
    return ret_y



"""
    Optimizer.BaselineOptimizer.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import logging
from scipy.optimize import basinhopping
LOG_IDENTIFICATION = "Baseline Optimizer"

class BaselineOptimizer:
    def __init__(self, optimizer):
        self.logger = logging.getLogger(__name__)
        self.optimizer = optimizer
        self.logger.info("%s constructed", LOG_IDENTIFICATION)

    def solve(self, init_params, baseline_indeces, real_bounds=None):
        self.indeces = baseline_indeces
        self.logger.info("strategic solve started with init_params=%s", str(init_params))
        self.optimizer.prepare_for_optimization(init_params, real_bounds=real_bounds)
        self.optimizer.cb_fh = self.cb_fh
        self.optimizer.write_init_callback_txt(init_params)

        self.temp_params = self.optimizer.to_norm_params(init_params)

        bounds = [(0, 10)]*len(self.indeces)
        minimizer_kwargs = dict(method='Nelder-Mead', bounds=bounds)
        result = basinhopping(self.baseline_objective, self.temp_params[self.indeces],
                              callback=self.minima_callback,
                              niter=20,
                              minimizer_kwargs=minimizer_kwargs)

        self.cb_fh.close()
        self.temp_params[self.indeces] = result.x
        result.x = self.temp_params
        return result
    
    def baseline_objective(self, p):
        pass
    
    def minima_callback(self, x, f, accept):
        self.temp_params[self.indeces] = x
        return self.optimizer.minima_callback(self.temp_params, f, accept)

def test_optimizer(caller):
    print("test_optimizer")
    optimizer = caller.get_optimizer()
    init_params = caller.get_init_params()
    optimizer.objective_func(init_params, plot=True)


"""
    Batch.BatchMappingImpl.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.stats import linregress
from scipy.optimize import minimize
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Models.ModelUtils import compute_cy_list, compute_area_props
from molass_legacy.QuickAnalysis.ModeledPeaks import get_modeled_peaks_impl
from molass_legacy.Peaks.PeakParamsSet import PeakParamsSet

def get_mapped_peaks(model, src_peaks, src_mapping_indeces, tgt_x, tgt_y, tgt_mapping_peaks):
    X = src_peaks[src_mapping_indeces,1]
    Y = tgt_mapping_peaks[:,1]
    slope, intercept = linregress(X, Y)[0:2]
    scale = np.average(tgt_mapping_peaks[:,0]/src_peaks[src_mapping_indeces,1])
    target_params = []
    for h, m, s, t, a in src_peaks:
        h_ = h*scale
        m_ = m*slope + intercept
        s_ = s*slope
        t_ = t*slope
        a_ = a*slope
        target_params.append((h_, m_, s_, t_, a_))
    target_params = np.array(target_params)
    temp_params = target_params.copy()

    def scale_objective(scales):
        cy_list = []
        for params, s in zip(temp_params, scales):
            params[0] = s
            cy = model(tgt_x, params)
            cy_list.append(cy)
        ty = np.sum(cy_list, axis=0)
        return np.sum((ty - tgt_y)**2)

    res = minimize(scale_objective, temp_params[:,0])
    temp_params[:,0] = res.x

    max_sigma = np.max(temp_params[:,2])*2
    max_tau = np.max(temp_params[:,3])*2

    def full_objective(p):
        params_array = p.reshape(temp_params.shape)
        cy_list = []
        for params in params_array:
            cy = model(tgt_x, params)
            cy_list.append(cy)
        ty = np.sum(cy_list, axis=0)

        max_sigma_ = np.max(params_array[:,2])
        max_tau_ = np.max(params_array[:,3])

        # penalty: to avoid too large sigma and tau
        penalty = max(0, max_sigma_ - max_sigma)**2 + max(0, max_tau_ - max_tau)**2

        return np.log(np.sum((ty - tgt_y)**2)) + np.log(1+penalty)*10

    res = minimize(full_objective, temp_params.flatten())

    return res.x.reshape(target_params.shape)

def show_mapped_result_impl(self, uv_x, uv_y, xr_x, xr_y, debug=False):
    mapping_src = self.mapping_src
    x = mapping_src.xr_x
    model = mapping_src.model
    cy_list = compute_cy_list(model, x, mapping_src.xr_peaks)
    props = compute_area_props(cy_list)
    print("props=", props)
    num_mainpeaks = 2
    indeces = np.argpartition(props, -num_mainpeaks)[-num_mainpeaks:]
    print(indeces)

    a, b = self.get_pre_recog_mapping_params()
    uv_peaks, xr_peaks = get_modeled_peaks_impl(a, b, uv_x, uv_y, xr_x, xr_y, 2, exact_num_peaks=2, affine=True)
    tgt_uv_peaks = get_mapped_peaks(model, mapping_src.uv_peaks, indeces, uv_x, uv_y, uv_peaks)
    tgt_xr_peaks = get_mapped_peaks(model, mapping_src.xr_peaks, indeces, xr_x, xr_y, xr_peaks)

    if debug:
        with plt.Dp():
            fig, axes = plt.subplots(nrows=2, ncols=2, figsize=(12, 10))
            fig.suptitle("Mapped Decomposition Result", fontsize=20)
            ax1, ax2 = axes[0,:]
            ax1.set_title("20220716/OA_ALD_201 (UV)", fontsize=16)
            ax2.set_title("20220716/OA_ALD_201 (XR)", fontsize=16)
            mapping_src.draw(axes_info=(ax1, ax2))
            ax1, ax2 = axes[1,:]
            ax1.set_title("20220716/OA_ALD_202 (UV)", fontsize=16)
            ax2.set_title("20220716/OA_ALD_202 (XR)", fontsize=16)
            ax1.plot(uv_x, uv_y, color='blue')
            for p in tgt_uv_peaks:
                cy = model(uv_x, p)
                ax1.plot(uv_x, cy, ":")
            ax2.plot(xr_x, xr_y, color='orange')
            for p in tgt_xr_peaks:
                cy = model(xr_x, p)
                ax2.plot(xr_x, cy, ":")
            fig.tight_layout()
            plt.show()

    self.peak_params_set = PeakParamsSet(tgt_uv_peaks, tgt_xr_peaks, a, b) 
    return tgt_uv_peaks, tgt_xr_peaks


"""
    CommandMain.py

    Copyright (c) 2016-2018, Masatsuyo Takahashi, KEK-PF
"""
import sys
import warnings
import traceback
import numpy                as np
from CommandLineOptions     import parser
# from KekToolsGP             import AutorgKek
# import KekLib, molass_legacyGuinierAnalyzer, SerialAnalyzer, Decomposer
from molass_legacy.GuinierAnalyzer.AutorgKekAdapter import AutorgKekAdapter as AutorgKek
from ResultPrinter          import ResultPrinter

DEBUG = False

def autorg_func( in_file, out_file, ASTSAS, robust ):
    if DEBUG: print( 'autorg_func: ', ASTSAS, robust )

    command_ = CommandMain()
    atsas_  = ASTSAS == 1
    robust_ = robust == 1

    try:
        result = command_.execute( in_file, out_file, ATSAS=atsas_, robust=robust_ )
    except Exception as e:
        if True:
            ( exc_type_type_, exc_value, exc_traceback ) = sys.exc_info()
            e_seq = traceback.format_exception( exc_type_type_, exc_value, exc_traceback )
            print( ''.join( e_seq ) )
        print( e )
        return -1

    if DEBUG: print( 'autorg_func: before return' )

    if robust_:
        return 0
    else:
        return -1 if result is None or result.Rg is None else 0

class CommandMain:
    def __init__( self ):
        self.opts = None
        warnings.filterwarnings("ignore")

    def run( self ):
        if self.opts is None:
            self.opts, self.args = parser.parse_args()

        if self.opts.version:
            from molass_legacy.AutorgKek.AppVersion import get_com_version_string
            print( get_com_version_string() )
            sys.exit()

        if self.opts.server:
            self.run_as_a_server()
            sys.exit()

        if len( self.args ) == 0:
            parser.print_help()
            sys.exit()

        if self.opts.output is None:
            self.fh = sys.stdout
        else:
            self.fh = open( self.opts.output, "w" )

        self.csv_output = ( self.opts.format == 'csv' or self.opts.output.find( '.csv' ) > 0 )

        self.separator = ',' if self.csv_output else '\t'
        self.ATSAS  = self.opts.ATSAS
        self.format = self.opts.format

        outfile = self.opts.output
        printer = ResultPrinter(
            fh = self.fh,
            csv_output = outfile is not None and outfile.find( '.csv' ) > 0,
            ATSAS=self.ATSAS,
            )

        multiple = len( self.args ) > 1
        for i, file in enumerate( self.args ):
            # print( file )
            autorg_ = AutorgKek( file )

            try:
                result = autorg_.run( robust=self.opts.robust )
            except Exception as e:
                sys.exit( str(e) )

            if not self.opts.robust and ( result is None or result.Rg is None ):
                sys.exit( "Can't determine Rg." )

            printer.print_result( i, file, result, multiple=multiple )

        if self.fh != sys.stdout:
            self.fh.close()

    def execute( self, in_file, out_file, ATSAS=False, robust=False, optimize=True ):
        self.autorg = AutorgKek( in_file )
        result = self.autorg.run( robust=robust, optimize=optimize )

        if result is None: return result

        if robust or result.Rg is not None:
            fh = open( out_file, "w" )

            printer = ResultPrinter(
                fh = fh,
                csv_output = out_file.find( '.csv' ) > 0,
                ATSAS=ATSAS,
                )
            printer.print_result( 0, in_file, result )

            fh.close()

        return result

    def run_as_a_server( self ):
        from ClientServer   import AutorgKekServer
        server = AutorgKekServer( self )
        server.run()



"""
    Optimizer.FullOptUtils.py

    Copyright (c) 2021-2024, SAXS Team, KEK-PF
"""
import re

def get_class_code_from_logfile(path):
    from .OptLogFile import OptLogFile
    log_file = OptLogFile(path)
    return log_file.class_code

def get_in_data_folder(in_data_info_txt):
    data_folder = None
    with open(in_data_info_txt) as fh:
        in_folder_re = re.compile("in_folder=(\S+)\s*")
        for line in fh:
            m = in_folder_re.match(line)
            if m:
                data_folder = m.group(1)
                break
    return data_folder

def in_folder_consistency_ok(in_folder, in_data_info_txt, parent):
    data_folder = get_in_data_folder(in_data_info_txt)

    folder_a = "/".join(in_folder.replace("\\", "/").split("/")[-2:])
    folder_b = "/".join(data_folder.replace("\\", "/").split("/")[-2:])
    ok = folder_a == folder_b
    if not ok:
        import molass_legacy.KekLib.CustomMessageBox as MessageBox
        yn = MessageBox.askyesno("Inconsistency Notification",
            "Input folders do not seem consistent as shown below.\n"
            "   Main Window Input: %s\n"
            "   Job  Result Input: %s\n"
            "The process can be unreliable.\n"
            "Would you like to proceed anyway?" % (folder_a, folder_b),
            parent=parent)
        if yn:
            ok = True
    return ok



# coding: utf-8
"""
    ODR_WLS.py

    Copyright (c) 2016, Masatsuyo Takahashi, KEK-PF
"""
import numpy                as np
from scipy                  import stats
from scipy.odr              import Model, Data, ODR
from LinearModel            import linear_model

class ODR_WLS_Result:
    def __init__( self, A, B, sigmaA, sigmaB ):
        self.params = np.array( [ A, B ] )
        self.std_errors = np.sqrt( [ sigmaA, sigmaB ] )
        self.sigmaA2    = sigmaA**2
        self.sigmaB2    = sigmaB**2

    def cov_params( self ):
        # this is diagonal only for proof purpose
        return np.diag( [ self.sigmaA2, self.sigmaB2 ] )

class ODR_WLS:
    def __init__( self, y, x, w ):

        slope, intercept, r_value, p_value, std_err = stats.linregress( x, y )

        linear      = Model( linear_model )
        mydata      = Data( x, y, w )
        myodr       = ODR( mydata, linear, beta0=[ slope, intercept ] )
        myoutput    = myodr.run()

        self.B,      self.A         = myoutput.beta
        self.sigmaB, self.sigmaA    = myoutput.sd_beta

    def fit( self ):
        return ODR_WLS_Result( self.A, self.B, self.sigmaA, self.sigmaB )



# coding: utf-8
"""
    LogicalExpression.py

    Copyright (c) 2021, Masatsuyo Takahashi, KEK-PF
"""

import sys
import re
import inspect
import ast

expr_re = re.compile(r"return\s+(.+)?\s+$")

class PrintNodeVisitor(ast.NodeVisitor):
    def visit(self, node):
        print(node)
        return super().visit(node)

class LogicalExpression:
    def __init__(self, closure):
        self.closure = closure
        self.structure = self.analyze(closure)

    def analyze(self, closure):
        src = inspect.getsource(closure)
        print(src)
        m = expr_re.search(src)
        if m:
            expr_src = m.group(1)
        else:
            assert False

        print("expr_src=", expr_src)
        expr = ast.parse(expr_src, mode='eval')
        print(ast.dump(expr, indent=4))

        visitor = PrintNodeVisitor()
        visitor.visit(expr)

    def tracer(self, *args):
        print("---- tracer:", *args)

    def evaluate(self):
        sys.settrace(self.tracer)
        ret = self.closure()
        sys.settrace(None)
        why = self.get_reason(ret)
        return ret, why

    def get_reason(self, judge):
        return None

def spike():
    pass



"""
    Models.Stochastic.OligoporeUtils.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize, basinhopping
from matplotlib.widgets import Slider, Button
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.KekLib.ExceptionTracebacker import log_exception
from molass_legacy.Models.Stochastic.ParamLimits import USE_K, M1_WEIGHT, M2_WEIGHT, BASINHOPPING_SCALE
from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder

def draw_exclusion_cuve_oligopore(axt, split_params, olgp_params, trs, rgs):
    # this exclusion cuve is not yet implement oligopore
    (N, T, x0), pszv, pszp, p_rgs, N0 = split_params(olgp_params)
    me, mp = 1.5, 1.5
    max_poresize = np.max(pszv)
    rv = np.linspace(max_poresize, 10, 100)
    rhov = rv/max_poresize
    tv = x0 + N*T*(1 - rhov)**(me+mp)
    axt.plot(tv, rv, lw=2, color="yellow")
    axt.plot(trs, rgs, "o", color="red")
    axt.axvline(x=x0, color="red")

def plot_oligopore_moments_state(x, y, cy_list, ty, egh_moments_list, moments, split_params, olgp_params, rgs, params_scaler, title=None, save_fig_as=None):
    top_params, pszv, pszp, p_rgs, N0 = split_params(olgp_params)

    print("pszv=", pszv, "pszp=", pszp)

    plot_params = np.asarray(olgp_params).copy()
    print("plot_params=", plot_params)

    slider_specs = [    ("N", 0, 6000, plot_params[0]),
                        ("T", 0, 3, plot_params[1]),
                        ("t0",  -1000, 1000, plot_params[2]),
                        ("me", 0, 3, plot_params[3]),
                        ("mp", 0, 3, plot_params[4]),
                        ("poresize", 20, 400, plot_params[5]),
                        ]

    with plt.Dp():
        fig, (ax1,ax2) = plt.subplots(ncols=2, figsize=(12, 5))
        fig.suptitle(title)
        ax1.set_title("Elution Curve")
        ax1.plot(x, y, color='orange')
        for k, (cy, M, eghM) in enumerate(zip(cy_list, moments, egh_moments_list)):
            ax1.plot(x, cy, ':')
            ax1.axvline(x=M[0], color="green")
            ax1.axvline(x=eghM[0], ls=":", color="cyan")
        bounds = params_scaler.get_bounds()
        t0_upper = bounds[2][1]
        ax1.axvline(x=t0_upper, color="pink")
        # ax1.plot(x, ty, ':', color="red")
        axt = ax1.twinx()
        axt.grid(False)
        trs = [M[0] for M in moments]
        draw_exclusion_cuve_oligopore(axt, split_params, olgp_params, trs, rgs)
        ax2.set_title("Pore Size Distribution")
        ax2.bar(pszv, pszp, width=10)
        fig.tight_layout()
        ret = plt.show()

def plot_oligopore_conformance_impl(ax, axt, N, K, t0, pszs, pszp, plot_info,
                                    rg_info=None, confrg_info=None,
                                    plot_cuvatures=False, ax2=None):

    if rg_info is None:
        max_rg = 120
    else:
        max_rg = max(120, np.max(rg_info[0]) * 1.1)
    rv = np.linspace(10, max_rg, 100)

    trs_list = []
    for poresize, p in zip(pszs, pszp):    
        rhov = rv/poresize
        rhov[rhov > 1] = 1
        trs_list.append(p * K*np.power(1 - rhov,3))
    tv = t0 + np.sum(trs_list, axis=0)
    x = plot_info.x
    y = plot_info.y
    model = plot_info.model
    cy_list = []
    for params in plot_info.peaks:
        cy = model(x, params)
        cy_list.append(cy)

    ax.set_title("Exclusion Points of %s against Exclusion Curve" % get_in_folder(), fontsize=16)
    ax.set_xlabel("Time (Frames)")
    ax.set_ylabel("Intensity")
    ax.plot(x, y, color="orange")
    for cy in cy_list:
        ax.plot(x, cy, ":")

    ax.axvline(x=t0, color="red", label="$t_0$=%d" % int(t0))
    ax.legend(loc="upper left")

    # axt = ax.twinx()
    axt.grid(False)
    axt.set_ylabel(r"$R_g (\AA_)$")
    def join_values(values):
        return ','.join(["%.4g" % v for v in values])
    PSD = "(%s), (%s)" % (join_values(pszs), join_values(pszp))
    axt.plot(tv, rv, color='yellow', lw=3, label="Estimated Exclusion Curve: K=%.4g, PSD=%s" % (K, PSD))
    ry = None
    if rg_info is not None:
        rgs, trs = rg_info[0:2]
        axt.plot(trs, rgs, "o")
        if len(rg_info) > 2:
            from scipy.interpolate import UnivariateSpline
            try:
                if rg_info[2]:  # draw spline
                    spline = UnivariateSpline(trs, rgs)
                ry = spline(tv)
                moderate = ry < 100     # avoid distortion caused by drawing extreme values as in 20201211
                axt.plot(tv[moderate], ry[moderate], color="cyan", alpha=0.5, lw=3, label="Spline of Exclusion Points")
            except:
                log_exception(None, "UnivariateSpline ry failure: ")
            
    if confrg_info is not None:
        rgs, trs = confrg_info
        axt.plot(trs, rgs, "o")

    ymin, ymax = axt.get_ylim()
    axt.set_ylim(max(0, ymin), max_rg)

    axt.legend(loc="upper right")

    if plot_cuvatures:
        from molass_legacy.KekLib.OurCurvature import curvature_curve
        ax2.set_title("Corresponding Curvatures", fontsize=16)
        ax2.set_xlabel("Time (Frames)")
        ax2.set_ylabel("Curvature")

        # to avoid ValueError: x must be strictly increasing if s = 0
        t = np.flip(tv)     
        diff_t = np.diff(t)
        diff_t_pos = np.where(diff_t > 0)[0]
        p = diff_t_pos[0]
        t_ = t[p:]
        try:
            cy = curvature_curve(t_, np.flip(rv)[p:])

            # to avoid anomalies
            m = np.argmax(cy)
            # there can be anomalies beyond m as in 20191006_OA_Ald_Ferr
            diff_cy = np.diff(cy[m:])
            diff_cy_pos = np.where(diff_cy > 0)[0]
            if len(diff_cy_pos) > 0:
                m_ = m + diff_cy_pos[-1] + 1
            else:
                # as in 20210727/data02 
                m_ = m

            ax2.plot(t_[m_:], cy[m_:], color="yellow", lw=3, label="Estimated Exclusion Curve Curvature")
        except:                
            log_exception(None, "curvature_curve rv failure: ")
        if ry is not None:
            try:
                cy_ = curvature_curve(t_, np.flip(ry)[p:])
                ax2.plot(t_, cy_, color="cyan", alpha=0.5, lw=3, label="Spline Curvature of Exclusion Points")
            except:
                log_exception(None, "curvature_curve ry failure: ")
        ax2.legend()

def plot_oligopore_conformance(N, K, t0, pszs, pszp, plot_info, rg_info=None, confrg_info=None, plot_cuvatures=False):
    from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder

    with plt.Dp():
        if plot_cuvatures:
            fig, (ax,ax2) = plt.subplots(ncols=2, figsize=(18,5))
        else:
            fig, ax = plt.subplots(figsize=(9,5))
            ax2 = None

        fig.suptitle("Exclusion Curve Estimate on %s" % get_in_folder(), fontsize=20)

        axt = ax.twinx()
        plot_oligopore_conformance_impl(ax, axt, N, K, t0, pszs, pszp, plot_info,
                                    rg_info=rg_info, confrg_info=confrg_info,
                                    plot_cuvatures=plot_cuvatures, ax2=ax2)

        fig.tight_layout()
        ret = plt.show()

    return ret

class OligoParamSplitter:
    def __init__(self, num_pszs, num_unreliables=0, dispersive=False):
        assert num_unreliables == 0
        assert not dispersive
        self.num_pszs = num_pszs
        self.num_unreliables = num_unreliables

    def __call__(self, p):
        num_pszs = self.num_pszs
        N, K, t0 = p[0:3]
        pszs = p[3:3+num_pszs]
        pszp = np.zeros(num_pszs)
        pszp[0:-1] = p[3+num_pszs:3+2*num_pszs-1]
        pszp[-1] = 1 - np.sum(pszp[0:-1])
        return N, K, t0, pszs, pszp

def guess_oligopore_colparams(num_pszs, rgs, qualities, egh_moments_list, init_params, m_props, debug=False, plot_info=None):
    from molass_legacy.Models.Stochastic.ParamLimits import OligoporeParamsScaler

    print("guess_oligopore_colparams: init_params=",init_params)

    assert USE_K    # support only USE_K

    split = OligoParamSplitter(num_pszs)

    me = 1.5
    mp = 1.5

    use_timescale = True
    use_basinhopping = True
    num_timescales = 1 if use_timescale else 0
    ORDER_PENALTY_SCALE = 1e6
    params_scaler = OligoporeParamsScaler(num_pszs, egh_moments_list, 0, rgs, num_timescales=num_timescales, allow_near_t0=True, t0_lower=None)    # t0_lower=700 for 20220716/FER_OA_302

    def objective(p, debug=False):
        if use_basinhopping:
            p = params_scaler.scale_back(p)
        N, K, t0, pszv, pszp = split(p)
        T = K/N
        if use_timescale:
            T *= p[-1]
            t0 *= p[-1]

        order_penalty = abs(min(0, np.min(np.diff(pszv)))) * ORDER_PENALTY_SCALE

        dev_list = []
        for k, (M, rg) in enumerate(zip(egh_moments_list, rgs)):
            rhov = rg/pszv
            rhov[rhov > 1] = 1
            M1_ = t0 + N * T * np.sum(pszp * (1 - rhov)**(me + mp))
            M2_ = 2 * N * T**2 * np.sum(pszp * (1 - rhov)**(me + 2*mp))
            dev_list.append(M1_WEIGHT*(M1_ - M[0])**2 + M2_WEIGHT*(M2_ - M[1])**2)

        return np.sum(np.asarray(dev_list) * m_props) + order_penalty

    bounds = params_scaler.get_bounds()
    print("bounds=", bounds)
    timescale = [1] if use_timescale else []
    init_params = np.concatenate([init_params, timescale])

    if use_basinhopping:
        init_params_ = params_scaler.scale(init_params)
        bounds_ = [(0, BASINHOPPING_SCALE)] * len(init_params)
        minimizer_kwargs = dict(method='Nelder-Mead', bounds=bounds_)
        res = basinhopping(objective, init_params_, minimizer_kwargs=minimizer_kwargs)
        res.x = params_scaler.scale_back(res.x)     # do not forget to scale back the result
    else:
        res = minimize(objective, init_params, method='Nelder-Mead', bounds=bounds)

    N, K, t0, pszs, pszp = split(res.x)
    if use_timescale:
        K *= res.x[-1]
        t0 *= res.x[-1]

    if debug and plot_info is not None:
        print("res.x=", res.x)
        trs = np.array([M[0] for M in egh_moments_list])
        ret = plot_oligopore_conformance(N, K, t0, pszs, pszp, plot_info, rg_info=(rgs, trs, True), plot_cuvatures=False)
        if not ret:
            return

    # return excluding time scale
    return N, K, t0, pszs, pszp


"""
    ScrolledFrame.py

    original: http://tkinter.unpythonic.net/wiki/VerticalScrolledFrame

    Copyright (c) 2016-2024, Masatsuyo Takahashi, KEK-PF
"""
import tkinter      as Tk
import tkinter.constants    as Tkconstants
from molass_legacy.KekLib.TkSupplements  import SlimButton

class ScrolledFrame(Tk.Frame):
    """A pure Tkinter scrollable frame that actually works!
    * Use the 'interior' attribute to place widgets inside the scrollable frame
    * Construct and pack/place/grid normally
    * This frame only allows vertical scrolling

    """
    def __init__(self, parent, *args, **kw):
        self.sync_canvas_info = None
        Tk.Frame.__init__(self, parent, *args, **kw)            

        self.rowconfigure(0, weight=1)
        self.columnconfigure(0, weight=1)

        # create a canvas object and a vertical scrollbar for scrolling it
        vscrollbar = Tk.Scrollbar(self, orient=Tk.VERTICAL)
        vscrollbar.grid( row=0, column=1, sticky=Tkconstants.NS )
        hscrollbar = Tk.Scrollbar(self, orient=Tk.HORIZONTAL)
        hscrollbar.grid( row=1, column=0, sticky=Tkconstants.EW )

        self.canvas = canvas = Tk.Canvas( self, bd=0, highlightthickness=0,
                            xscrollcommand=hscrollbar.set,
                            yscrollcommand=vscrollbar.set,
                            )
        canvas.grid( row=0, column=0, sticky=Tkconstants.NSEW )

        self.canvas_bottom = canvas_bottom = Tk.Canvas( self, bd=0, highlightthickness=0,
                            xscrollcommand=hscrollbar.set,
                            height=0
                            )

        canvas_bottom.grid( row=2, column=0, sticky=Tkconstants.NSEW )

        def xview_( *args ):
            # print( 'xview_' )
            canvas.xview( *args )
            canvas_bottom.xview( *args )
            if self.sync_canvas_info is not None:
                for sync_canvas in self.sync_canvas_info:
                    sync_canvas.xview( *args )

        hscrollbar.config( command=xview_ )
        vscrollbar.config( command=canvas.yview )

        # reset the view
        canvas.xview_moveto(0)
        canvas.yview_moveto(0)
        canvas_bottom.xview_moveto(0)

        self.canvas_info = canvas, canvas_bottom

        # create a frame inside the canvas which will be scrolled with it
        self.interior = interior = Tk.Frame(canvas)
        interior_id = canvas.create_window(0, 0, window=interior,
                                           anchor=Tk.NW)

        self.interior_bottom = interior_bottom = Tk.Frame(canvas_bottom)
        interior_bottom_id = canvas_bottom.create_window(0, 0, window=interior_bottom,
                                           anchor=Tk.NW)

        # track changes to the canvas and frame width and sync them,
        # also updating the scrollbar
        def _configure_interior( canvas, interior, event ):
            # update the scrollbars to match the size of the inner frame
            canvas.config( scrollregion=( 0, 0, interior.winfo_reqwidth(), interior.winfo_reqheight() ) )
            if interior.winfo_reqwidth() != canvas.winfo_width():
                # update the canvas's width to fit the inner frame
                canvas.config( width=interior.winfo_reqwidth() )
            if interior.winfo_reqheight() != canvas.winfo_height():
                # update the canvas's height to fit the inner frame
                canvas.config( height=interior.winfo_reqheight() )

        interior.bind('<Configure>', lambda event: _configure_interior( canvas, interior, event ) )
        interior_bottom.bind('<Configure>', lambda event: _configure_interior( canvas_bottom, interior_bottom, event ))

        def _configure_canvas( canvas, interior, interior_id, event, height_adjust=True ):
            # TODO: is this neccesary?
            if interior.winfo_reqwidth() != canvas.winfo_width():
                # update the inner frame's width to fill the canvas
                canvas.itemconfigure( interior_id, width=canvas.winfo_width() )
            if height_adjust and interior.winfo_reqheight() != canvas.winfo_height():
                # update the inner frame's heght to fill the canvas
                canvas.itemconfigure( interior_id, height=canvas.winfo_height() )

        # canvas.bind('<Configure>', lambda event: _configure_canvas( canvas, interior, interior_id, event ) )
        # interior_bottom.bind('<Configure>', lambda event: _configure_canvas( canvas_bottom, interior_bottom, interior_bottom_id, event, height_adjust=False ) )

        self.lower_right_button = SlimButton( self, text='▲', height=18, width=18, command=self.lower_right_button_toggle )
        self.lower_right_button.grid( row=1, column=1 )
        self.lower_right_button_toggle()

    # thanks to: https://www.daniweb.com/programming/software-development/code/429838/simple-tkinter-toggle-button
    def lower_right_button_toggle( self ):
        text_ = self.lower_right_button.button['text']
        if text_ == '▼':
            text_   = '▲'
            self.canvas_bottom.grid( row=2, column=0, sticky=Tkconstants.NSEW  )
        else:
            text_  = '▼'
            self.canvas_bottom.grid_remove()
        self.lower_right_button.button['text'] = text_

    def destroy( self ):
        self.canvas.destroy()   # seems to have to be destroyed first in Python 3.7
        super(Tk.Frame, self).destroy()

    # thanks to: https://stackoverflow.com/questions/17355902/tkinter-binding-mousewheel-to-scrollbar
    def add_bind_mousewheel(self):
        self.canvas.bind_all("<MouseWheel>", self._on_mousewheel)

    def _on_mousewheel(self, event):
        self.canvas.yview_scroll(int(-1*(event.delta/120)), "units")

    def set_sync_info(self, canvas_info):
        self.sync_canvas_info = canvas_info

if __name__ == "__main__":

    class SampleApp( Tk.Tk ):
        def __init__( self, *args, **kwargs ):
            root = Tk.Tk.__init__( self, *args, **kwargs )

            quit_button = Tk.Button( root, text="quit", command=self.quit )
            quit_button.pack( side=Tk.BOTTOM )

            self.frame = ScrolledFrame( root )
            self.frame.pack( fill=Tk.BOTH, expand=1 )
            self.frame2 = ScrolledFrame( root )
            self.frame2.pack( fill=Tk.BOTH, expand=1 )
            self.label = Tk.Label(self, text="Shrink the window to activate the scrollbar.")
            self.label.pack()
            buttons = []
            for i in range(10):
                buttons.append(Tk.Button( self.frame.interior, width=40, text="Button " + str(i)))
                buttons[-1].pack()

            self.frame.interior_bottom.configure( height=40 )
            for j in range(2):
                b = Tk.Button( self.frame.interior_bottom, width=20, text="Button B " + str(j) )
                b.grid( row=0, column=j )

            label = Tk.Label(self.frame2.interior, text="Frame to be sychronized", width=20)
            label.pack()
            b = Tk.Button( self.frame2.interior, width=40, text="Button S ")
            b.pack()

            self.frame.set_sync_info( self.frame2.canvas_info )
            self.frame2.set_sync_info( self.frame.canvas_info )

    app = SampleApp()
    app.mainloop()


"""
    temporary investigation to a suspected bug of multiprocessing.shared_memory
"""

"""Provides shared memory for direct access across processes.

The API of this package is currently provisional. Refer to the
documentation for details.
"""

__all__ = [ 'SharedMemory' ]

from functools import partial
import mmap
import os
import errno
import struct
import secrets

import _winapi
_USE_POSIX = False
_O_CREX = os.O_CREAT | os.O_EXCL

# FreeBSD (and perhaps other BSDs) limit names to 14 characters.
_SHM_SAFE_NAME_LENGTH = 14

# Shared memory block name prefix
_SHM_NAME_PREFIX = 'wnsm_'

def _make_filename():
    "Create a random filename for the shared memory object."
    # number of random bytes to use for name
    nbytes = (_SHM_SAFE_NAME_LENGTH - len(_SHM_NAME_PREFIX)) // 2
    assert nbytes >= 2, '_SHM_NAME_PREFIX too long'
    name = _SHM_NAME_PREFIX + secrets.token_hex(nbytes)
    assert len(name) <= _SHM_SAFE_NAME_LENGTH
    return name


class SharedMemory:
    """Creates a new shared memory block or attaches to an existing
    shared memory block.

    Every shared memory block is assigned a unique name.  This enables
    one process to create a shared memory block with a particular name
    so that a different process can attach to that same shared memory
    block using that same name.

    As a resource for sharing data across processes, shared memory blocks
    may outlive the original process that created them.  When one process
    no longer needs access to a shared memory block that might still be
    needed by other processes, the close() method should be called.
    When a shared memory block is no longer needed by any process, the
    unlink() method should be called to ensure proper cleanup."""

    # Defaults; enables close() and unlink() to run without errors.
    _name = None
    _fd = -1
    _mmap = None
    _buf = None
    _flags = os.O_RDWR
    _mode = 0o600
    _prepend_leading_slash = True if _USE_POSIX else False

    def __init__(self, name=None, create=False, size=0):
        if not size >= 0:
            raise ValueError("'size' must be a positive integer")
        if create:
            self._flags = _O_CREX | os.O_RDWR
        if name is None and not self._flags & os.O_EXCL:
            raise ValueError("'name' can only be None if create=True")

        # Windows Named Shared Memory

        if create:
            while True:
                temp_name = _make_filename() if name is None else name
                # Create and reserve shared memory block with this name
                # until it can be attached to by mmap.
                h_map = _winapi.CreateFileMapping(
                    _winapi.INVALID_HANDLE_VALUE,
                    _winapi.NULL,
                    _winapi.PAGE_READWRITE,
                    (size >> 32) & 0xFFFFFFFF,
                    size & 0xFFFFFFFF,
                    temp_name
                )
                try:
                    last_error_code = _winapi.GetLastError()
                    if last_error_code == _winapi.ERROR_ALREADY_EXISTS:
                        if name is not None:
                            raise FileExistsError(
                                errno.EEXIST,
                                os.strerror(errno.EEXIST),
                                name,
                                _winapi.ERROR_ALREADY_EXISTS
                            )
                        else:
                            continue
                    self._mmap = mmap.mmap(-1, size, tagname=temp_name)
                finally:
                    _winapi.CloseHandle(h_map)
                self._name = temp_name
                break

        else:
            self._name = name
            # Dynamically determine the existing named shared memory
            # block's size which is likely a multiple of mmap.PAGESIZE.
            h_map = _winapi.OpenFileMapping(
                _winapi.FILE_MAP_READ,
                False,
                name
            )
            try:
                p_buf = _winapi.MapViewOfFile(
                    h_map,
                    _winapi.FILE_MAP_READ,
                    0,
                    0,
                    0
                )
            except:
                print('MapViewOfFile Error!')
            finally:
                print('h_map=', h_map)
                _winapi.CloseHandle(h_map)
            size = _winapi.VirtualQuerySize(p_buf)
            self._mmap = mmap.mmap(-1, size, tagname=name)

        self._size = size
        self._buf = memoryview(self._mmap)
        print(__name__, '__init__ ok')

    def __del__(self):
        try:
            self.close()
        except OSError:
            pass

    def __reduce__(self):
        return (
            self.__class__,
            (
                self.name,
                False,
                self.size,
            ),
        )

    def __repr__(self):
        return f'{self.__class__.__name__}({self.name!r}, size={self.size})'

    @property
    def buf(self):
        "A memoryview of contents of the shared memory block."
        return self._buf

    @property
    def name(self):
        "Unique name that identifies the shared memory block."
        reported_name = self._name
        if _USE_POSIX and self._prepend_leading_slash:
            if self._name.startswith("/"):
                reported_name = self._name[1:]
        return reported_name

    @property
    def size(self):
        "Size in bytes."
        return self._size

    def close(self):
        print(__name__, 'close')
        """Closes access to the shared memory from this instance but does
        not destroy the shared memory block."""
        if self._buf is not None:
            self._buf.release()
            self._buf = None
        if self._mmap is not None:
            self._mmap.close()
            self._mmap = None
        if _USE_POSIX and self._fd >= 0:
            os.close(self._fd)
            self._fd = -1
        print(__name__, 'close ok')

    def unlink(self):
        """Requests that the underlying shared memory block be destroyed.

        In order to ensure proper cleanup of resources, unlink should be
        called once (and only once) across all processes which have access
        to the shared memory block."""
        if _USE_POSIX and self._name:
            from .resource_tracker import unregister
            _posixshmem.shm_unlink(self._name)
            unregister(self._name, "shared_memory")



# coding: utf-8
"""
    LowPercentileSpline.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.interpolate import RectBivariateSpline
from BasePlane import LowPercentilePlane
import molass_legacy.KekLib.DebugPlot as plt
# import matplotlib.pyplot as plt
from molass_legacy.KekLib.OurMatplotlib import get_color

USE_OVERLAPPING_PLANES  = False
USE_AVERAGED_JOINTS     = False
USE_GEOMDL_FITTING      = False

def bspline_surface(control_points, num_planes, fig, axis_recs):
    # import matplotlib.pyplot as mplt
    if USE_GEOMDL_FITTING:
        from geomdl import fitting
    else:
        from geomdl import BSpline , utilities
    from geomdl.visualization.VisMPL import VisConfig
    from OurGeomdl import VisSurface

    if USE_OVERLAPPING_PLANES:
        n = num_planes + 2
    elif USE_AVERAGED_JOINTS:
        n = num_planes*2 + 1
    else:
        n = num_planes + 1

    if USE_GEOMDL_FITTING:
        size_u = n
        size_v = 3
        degree_u = 3
        degree_v = 2

        # Do global surface approximation
        surf = fitting.approximate_surface(control_points, size_u, size_v, degree_u, degree_v)
    else:
        # Create a surface instance
        surf = BSpline.Surface()

        # Set degrees
        surf.degree_u = 3
        surf.degree_v = 2

        # Set control points
        surf.set_ctrlpts(control_points, n, 3)

        # Auto-generate knot vectors
        surf.knotvector_u = utilities.generate_knot_vector(surf.degree_u, surf.ctrlpts_size_u)
        surf.knotvector_v = utilities.generate_knot_vector(surf.degree_v, surf.ctrlpts_size_v)

        # Set sample size
        # surf.sample_size = 25

        # Evaluate surface
        surf.evaluate()

    # Set the visualization component
    alpha = 0.5
    config = VisConfig(alpha=alpha)

    for ax, ax_lim in axis_recs:
        vis_component = VisSurface(config=config, fig=fig, ax=ax, alpha=alpha)
        surf.vis = vis_component

        # Plot the surface
        surf.render(plot=False)
        # axes = fig.get_axes()
        # print(len(axes))
        # ax = fig.gca()
        ax.set_xlim(ax_lim.get_xlim())
        ax.set_ylim(ax_lim.get_ylim())
        ax.set_zlim(ax_lim.get_zlim())
        # plt.show()

class LowPercentileSpline:
    def __init__(self, sd):
        array = sd.intensity_array[:,:,1].T

        full_width = 40
        half_width = full_width//2
        num_planes = 15

        q = sd.intensity_array[0,:,0]
        i = np.arange(array.shape[0])
        j = np.arange(array.shape[1])
        ii, jj = np.meshgrid(i, j)
        xx = q[ii]
        zz = array[ii,jj]

        smp_i = (sd.xray_slice.start + sd.xray_slice.stop)//2
        iy = slice(0, array.shape[1])

        control_points = []
        lpm_list = []
        abc_list = []
        for k in range(num_planes):
            if USE_OVERLAPPING_PLANES:
                start = k*half_width
                stop = (k+1)*full_width
            else:
                start = k*full_width
                stop = (k+1)*full_width

            data = array[start:stop,:]

            zpp = np.percentile(data, 30)
            wpp = np.where( data < zpp )
            x_ = q[start+wpp[0]]
            y_ = wpp[1]
            z_ = data[wpp[0], wpp[1]]
            lpm_list.append([x_, y_, z_])
            ix = slice(start, stop)
            lpp = LowPercentilePlane(array, q, sd.ivector, ix, iy, smp_i, debug=False)
            abc = lpp.get_params()
            abc_list.append([ix, abc])

            a, b, c = abc
            middle = (start + stop)//2

            if USE_OVERLAPPING_PLANES:
                if k == 0:
                    index = [start, middle]
                elif k == num_planes - 1:
                    index = [middle, stop]
                else:
                    index = [middle]

                for m, x in enumerate(q[index]):
                    for n, y in enumerate([iy.start, (iy.start+iy.stop)/2, iy.stop]):
                        z = a*x + b*y + c
                        point = [x, y, z]
                        control_points.append(point)
            elif USE_AVERAGED_JOINTS:
                index = [start, middle, stop]
                keep_points = []
                for m, x in enumerate(q[index]):
                    for n, y in enumerate([iy.start, (iy.start+iy.stop)/2, iy.stop]):
                        z = a*x + b*y + c
                        point = [x, y, z]
                        if m == 0:
                            if k == 0:
                                control_points.append(point)
                            else:
                                last_point = kept_points[n]
                                average_point = list(np.average([last_point, point], axis=0))
                                control_points.append(average_point)
                        elif m == 1:
                            control_points.append(point)
                        elif m == 2:
                            if k < num_planes - 1:
                                keep_points.append(point)
                            else:
                                control_points.append(point)
                kept_points = keep_points
            else:
                index = [start] if k < num_planes - 1 else [start, stop]
                for m, x in enumerate(q[index]):
                    for n, y in enumerate([iy.start, (iy.start+iy.stop)/2, iy.stop]):
                        z = a*x + b*y + c
                        point = [x, y, z]
                        control_points.append(point)

        debug = True
        if debug:
            fig = plt.figure(figsize=(24,14))
            ax1 = fig.add_subplot(231, projection='3d')
            ax2 = fig.add_subplot(232, projection='3d')
            ax3 = fig.add_subplot(233, projection='3d')
            ax4 = fig.add_subplot(234, projection='3d')
            ax5 = fig.add_subplot(235, projection='3d')
            ax6 = fig.add_subplot(236, projection='3d')

            ax1.plot_surface(xx, jj, zz)

            for k, rec in enumerate(lpm_list):
                x_, y_, z_ = rec

                ix, abc = abc_list[k]
                xs = q[ix]
                ys = np.arange( iy.start, iy.stop )
                xx, yy = np.meshgrid( np.linspace(xs[0], xs[-1], 10), np.linspace(ys[0], ys[-1], 10) )
                A, B, C = abc
                zz = xx * A + yy * B + C

                color = get_color(k)
                for ax in [ax1, ax4]:
                    ax.plot(x_, y_, z_, 'o', markersize=1, color=color)

                for ax in [ax2, ax3, ax5, ax6]:
                    ax.plot_surface(xx, yy, zz, color='red', alpha=0.1 )

            x_, y_, z_ = zip(*control_points)
            for ax in [ax2, ax5]:
                ax.plot(x_, y_, z_, 'o')

            ax2.set_xlim(ax1.get_xlim())
            ax2.set_ylim(ax1.get_ylim())
            ax2.set_zlim(ax1.get_zlim())
            ax4.set_xlim(ax5.get_xlim())
            ax4.set_ylim(ax5.get_ylim())
            ax4.set_zlim(ax5.get_zlim())

            axis_recs = [[ax3, ax1], [ax6, ax5]]
            bspline_surface(control_points, num_planes, fig, axis_recs)

            fig.tight_layout()
            plt.show()
            plt.close()



"""
    Models.Stochastic.RgReliability.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import logging
import numpy as np

def determine_unreliables(rgs, qualities, props, trust_max_num=None, debug=True):
    logger = logging.getLogger(__name__)
    logger.info("determin_unreliables: rgs=%s, props=%s, qualities=%s, trust_max_num=%s", rgs, props, qualities, trust_max_num)

    rg_diff = np.diff(rgs)
    desc_indeces = np.where(rg_diff > 0)[0]
    indeces_from_desc_test = []
    for k in desc_indeces:
        # pick one with a lower quality rg as unreliable if they are not in decending order
        indeces_from_desc_test.append(k if qualities[k] < qualities[k+1] else k+1)

    indeces_from_quality_test = np.where(np.logical_or(props < 0.05,
                                                       qualities < 0.3
                                                       ))[0]

    ret_indeces = np.asarray(list(set(indeces_from_desc_test) | set(indeces_from_quality_test)), dtype=int)   # dtype=int is required when it is empty
    if trust_max_num is not None:
        num_peaks = len(rgs)
        if num_peaks - len(ret_indeces) > trust_max_num:
            full_indeces = np.arange(num_peaks)
            trusted_indeces = np.setdiff1d(full_indeces, ret_indeces)
            trustes_qualities = qualities[trusted_indeces]
            num_remove = len(trusted_indeces) - trust_max_num
            kk = np.argpartition(trustes_qualities, num_remove)
            if debug:
                print("qualities=", qualities, "ret_indeces=", ret_indeces)
                print("trusted_indeces=", trusted_indeces)
                print("trustes_qualities=", trustes_qualities, "kk=", kk, "remove=", kk[:num_remove])
            ret_indeces = np.asarray(sorted(np.concatenate([ret_indeces, trusted_indeces[kk[:num_remove]]])))
    if debug:
        print("indeces_from_desc_test=", indeces_from_desc_test)
        print("indeces_from_quality_test=", indeces_from_quality_test)
        print("ret_indeces=", ret_indeces)
    
    return ret_indeces


"""
    MappingV2Utils.py

    Copyright (c) 2022-2023, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy._MOLASS.SerialSettings import get_setting
from .MappingParams import MappedInfo

def make_mapped_info_for_v2(x_curve):
    # this function separates v2 from the old v1 mapping

    unified_baseline_type = get_setting("unified_baseline_type")
    if unified_baseline_type == 1:
        xray_baseline_type = 1          # linear
    elif unified_baseline_type in [2, 3]:
        xray_baseline_type = 5          # integral
    else:
        assert False

    # affine_info
    # see ElutionMapper.get_affine_info()
    x = x_curve.x
    x_base = np.zeros(len(x))
    x_base_adjustment = 0
    peaks   = [ info[1] for info in x_curve.peak_info ]
    peak_x  = np.average( peaks )
    peak_y  = np.average( x_curve.spline( peaks ) )
    affine_info = [ x, x_base, x_base_adjustment, (peak_x, peak_y), peaks ]

    # see also  ElutionMapper.prepare_env_for_plain_LPM(self):
    opt_params = {  'xray_baseline_opt':1,
                    'xray_baseline_type':xray_baseline_type,
                    'xray_baseline_adjust':0,
                    'xray_baseline_with_bpa':1}

    return MappedInfo(opt_params=opt_params, affine_info=affine_info, x_curve=x_curve, x_base=x_base)



# coding: utf-8
"""

    MenuButton.py

    Copyright (c) 2019-2022, SAXS Team, KEK-PF

"""
from molass_legacy.KekLib.OurTkinter import Tk

class MenuButton(Tk.Button):
    def __init__(self, parent, text, command_list=[]):
        self.parent = parent
        Tk.Button.__init__(self, parent, text=text, command=self.popup_menu)
        self.create_menu(command_list)

    def create_menu(self, command_list):
        self.popup_menu = Tk.Menu(self, tearoff=0)
        for label, command in command_list:
            self.popup_menu.add_command(label=label, command=command)

    def popup_menu(self):
        x = self.winfo_rootx()
        y = self.winfo_rooty()
        h = self.winfo_height()
        self.popup_menu.post(x, y+h)

    def entryconfig(self, *args, **kwargs):
        self.popup_menu.entryconfig(*args, **kwargs)



"""
    MplAnnotate.py

    * [How do I find the largest empty space in such images?]
        (https://stackoverflow.com/questions/67421025/how-do-i-find-the-largest-empty-space-in-such-images)
    * [How to read image from in memory buffer (StringIO) or from url with opencv python library]
        (https://stackoverflow.com/questions/13329445/how-to-read-image-from-in-memory-buffer-stringio-or-from-url-with-opencv-pytho)

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import numpy as np
import cv2 as cv

def get_opencv_img_from_buffer(buffer):
    bytes_as_np_array = np.frombuffer(buffer.read(), dtype=np.uint8)
    return cv.imdecode(bytes_as_np_array, flags=cv.IMREAD_UNCHANGED)   # 

def get_image_from_axis(fig, ax):
    import io
    import matplotlib.image as mpimg
    extent = ax.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
    buffer = io.BytesIO()
    fig.savefig(buffer, bbox_inches=extent)
    buffer.seek(0)
    # img = mpimg.imread(buffer)
    # img = cv.imread(buffer)
    img = get_opencv_img_from_buffer(buffer)
    return img

MARGIN_ADJUST_RATIO = 0.1

def get_annotate_position(ax, debug=False):
    fig = ax.figure
    color_img = get_image_from_axis(fig, ax)
    gray_img = cv.cvtColor(color_img, cv.COLOR_BGR2GRAY)
    kernel = np.ones((5,5),np.uint8)
    erosion = cv.erode(gray_img, kernel, iterations=1)
    thresh = cv.threshold(erosion, 128, 255, cv.THRESH_BINARY)[1]

    # add black border around threshold image to avoid corner being largest distance
    thresh2 = cv.copyMakeBorder(thresh, 1,1,1,1, cv.BORDER_CONSTANT, (0))
    distimg = cv.distanceTransform(thresh2, cv.DIST_L2, 5)

    min_val, max_val, min_loc, max_loc = cv.minMaxLoc(distimg)
    centx = max_loc[0] - 1  # -1 : adjust by border width
    centy = max_loc[1] - 1  # -1 : adjust by border width
    radius = int(max_val)
    isize, jsize = gray_img.shape   # note that image shape (i, j) corresponds to (y, x)
    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()

    if debug:
        import molass_legacy.KekLib.DebugPlot as dplt

        print("centx, centy, radius=", centx, centy, radius)
        print("xmin, xmax=", xmin, xmax)
        print("ymin, ymax=", ymin, ymax)

        with dplt.Dp():
            fig_, ax_ = dplt.subplots()
            ax.imshow(distimg)
            dplt.show()

    ret_values = []
    for vmin, vmax, size, v in [
        (xmin, xmax, jsize, centx),
        (ymin, ymax, isize, isize - centy),
        ]:
        w = v/size
        ret_value = vmin*(1 - w) + vmax*w
        ret_values.append(ret_value)

    src_scale = np.sqrt(isize**2 + jsize**2)
    tgt_scale = np.sqrt((xmax - xmin)**2 + (ymax - ymin)**2)
    ret_radius = radius/src_scale*tgt_scale

    return ret_values + [ret_radius]



"""
    UV.UvPreRecog.py

    Copyright (c) 2022-2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize, basinhopping
import logging
import molass_legacy.KekLib.DebugPlot as plt
from .PlainCurve import make_secondary_e_curve_at
from molass_legacy.Trimming.FlowChangeCandidates import get_largest_gradients
from molass_legacy.Trimming.Sigmoid import ex_sigmoid
from molass_legacy.Peaks.EghSupples import egh, d_egh
from molass_legacy.KekLib.SciPyCookbook import smooth
from molass_legacy.KekLib.BasicUtils import Struct

USE_SPLINE_MODEL = True

class UvPreRecog:
    def __init__(self, sd, pre_recog, debug=False, fig_file=None):
        if USE_SPLINE_MODEL:
            if debug:
                from importlib import reload
                import Baseline.UvBaseSpline
                reload(Baseline.UvBaseSpline)
            from molass_legacy.Baseline.UvBaseSpline import UvBaseSpline
        else:
            from molass_legacy.Baseline.UvBaseCurve import UvBaseCurve

        self.logger = logging.getLogger(__name__)

        fc = pre_recog.flowchange
        a_curve = fc.a_curve
        a_curve2 = fc.a_curve2

        x = a_curve2.x
        y = a_curve2.y

        fc_points = fc.get_real_flow_changes()
        fc_point = fc_points[0]

        pp_, ppe, ret_slice = fc.remove_irregular_points()

        if USE_SPLINE_MODEL:
            base_curve = UvBaseSpline(sd, a_curve)
            params = base_curve.guess_params(a_curve2, pp_, ret_slice, fc_point=fc_point, debug=debug)
        else:
            base_curve = UvBaseCurve(sd, a_curve, 1, 1)
            params = base_curve.guess_params(a_curve2, pp_, ret_slice, debug=debug)

        if debug:
            from molass_legacy.Baseline.Constants import SLOPE_SCALE
            from molass_legacy.UV.PlainCurveUtils import get_both_wavelengths

            exsig_params = params.copy()
            exsig_params[4:6] /= SLOPE_SCALE
            w1, w2 = get_both_wavelengths(sd.lvector)
            gy = fc.gy
            with plt.Dp():
                from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder
                from molass_legacy.Elution.CurveUtils import simple_plot
                from molass_legacy.DataStructure.MatrixData import simple_plot_3d
                data = sd.conc_array
                wv = sd.lvector

                wide_figure = False

                figsize = (20,5) if wide_figure else (20, 5)
                fig = plt.figure(figsize=figsize)
                if wide_figure:
                    ax1 = fig.add_subplot(141, projection="3d")
                    ax2 = fig.add_subplot(142)
                    ax3 = fig.add_subplot(143)
                    ax4 = fig.add_subplot(144)
                else:
                    ax2 = fig.add_subplot(141)
                    ax21 = fig.add_subplot(142)
                    ax22 = fig.add_subplot(143)
                    ax3 = fig.add_subplot(144)

                fig.suptitle("UV Baseline Model Fitting for %s" % get_in_folder(), fontsize=20)
                if wide_figure:
                    ax1.set_title("3D View", fontsize=16)
                ax2.set_title("Elution at λ=%.3g" % w1, fontsize=16)
                ax3.set_title("Elution at λ=%.3g" % w2, fontsize=16)
                if wide_figure:
                    ax4.set_title("Gradient at λ=%.3g" % w2, fontsize=16)
                else:
                    for ax in [ax21, ax22, ax3]:
                        ax.set_title("Elution at λ=%.3g" % w2, fontsize=16)

                if wide_figure:
                    simple_plot_3d(ax1, data, x=wv)

                simple_plot(ax2, a_curve)
                ax3.plot(x, y, label="data")
                ax3.plot(x, base_curve(x, params, y), label="baseline model")
                ax3.legend()
                if wide_figure:
                    ax4.plot(x, gy)
                    for k in pp_:
                        ax4.plot(x[k], gy[k], "o")
                    if len(ppe) > 0:
                        ax4.plot(x[ppe], gy[ppe], "o", color="red", label="irregular")
                    ax4.legend()
                else:
                    ax21.plot(x, y, label="data")
                    ax21.plot(x, ex_sigmoid(x, *exsig_params[0:6]), label="ex_sigmoid")
                    ax21.plot(x, base_curve.diff_spline(x), y, label="derivative")
                    ax21.legend()
                    ax22.plot(x, y, label="data")
                    ax22.plot(x, ex_sigmoid(x, *exsig_params[0:6]), label="ex_sigmoid")
                    ax22.plot(x, params[6]*base_curve.diff_spline(x), label="scaled derivative")
                    ymin, ymax = ax22.get_ylim()
                    tx = x[-1]*0.7
                    ty = (ymin + ymax)/2
                    # ty = 0.2*ymin + 0.8*ymax
                    ax22.text(tx, ty, "Scale=%.2g" % params[6], fontsize=16, alpha=0.5)
                    ax22.legend()
                fig.tight_layout()

                if fig_file is None:
                    plt.show()
                else:
                    from time import sleep
                    fig.savefig(fig_file)
                    plt.show(block=False)
                    sleep(0.5)

        self.trim_slice = ret_slice
        self.base_curve = base_curve
        self.init_params = params

    def remove_irregular_points(self, x, y, gy, pp3):
        ret_pp = []
        err_pp = []
        for k in pp3:
            if k < len(y) - 10:
                ret_pp.append(k)
            else:
                err_pp.append(k)
        if len(err_pp) > 0:
            ret_slice = slice(None, np.min(err_pp))
        else:
            ret_slice = slice(None, None)
        return np.array(sorted(ret_pp, key=lambda j: -abs(gy[j]))), np.array(err_pp), ret_slice

    def get_trim_slice(self):
        return self.trim_slice

    def get_base_curve_info(self):
        return self.base_curve, self.init_params



# DPHEM library which is compatible with sklearn.mixture.GaussianMixture 
# ---
# If you use this code, please cite our paper:
#   Lei Yu, Tianyu Yang, and Antoni B. Chan. "Density Preserving Hierarchical EM Algorithm: Simplifying 
#   Gaussian Mixture Models for Approximate Inference." TPAMI 2019
# ---
# Copyright (c) 2019-04-30
# Lei Yu, Antoni B. Chan
# Department of Computer Science
# City University of Hong Kong 
#
# v0.1

from matplotlib.patches import Ellipse
import matplotlib.pyplot as plt
import numpy as np
from sklearn.mixture import GaussianMixture



def do_split(ingmm,tagk,mytarget=None):
    '''gmm = do_split(ingmm,tagk)
       split ingmm to tagk components'''

    gmm = ingmm

    origk = ingmm.n_components
    if tagk > 2*origk:
        raise ValueError('Split target K too large.')

    myeigs = np.zeros(origk)
    for jj in range(origk):
        if not np.size(ingmm.covariances_[jj]):
            myeigs[jj] = -1
        else:
            if (ingmm.covariance_type == 'spherical' or
               ingmm.covariance_type == 'diag'):
               myeigs[jj] = max(ingmm.covariances_[jj])
            elif ingmm.covariance_type == 'full':
                D,V = np.linalg.eig(ingmm.covariances_[jj])
                myeigs[jj] = max(D)
            else:
                raise ValueError('Not Supported.')

    if mytarget is not None:
        dotarget = True
    else:
        dotarget = False

    while ingmm.n_components < tagk or dotarget:
        foo = max(myeigs)
        ind1 = myeigs.tolist().index(foo)

        # decide the target component
        if dotarget:
            ind2 = mytarget
            dotarget = False
        else:
            ind2 = ingmm.n_components
            gmm.n_components = ingmm.n_components + 1

        # make a new components
        tmp = ingmm.weights_[ind1]
        gmm.weights_[ind1] = tmp/2.0

        tmpcv = ingmm.covariances_[ind1]
        covariances = ingmm.covariances_.tolist().append(tmpcv)
        if ind2 in range(origk):
            gmm.weights_[ind2] = tmp/2.0
            gmm.covariances_[ind2] = tmpcv
        else:
            gmm.weights_ = np.asarray(np.append(ingmm.weights_,tmp/2.0))
            gmm.covariances_ = np.asarray(covariances)

        tmpmu = ingmm.means_[ind1]

        # perturb mu in maximum variance direction by std/2
        if ingmm.covariance_type == 'spherical':
            d = len(ingmm.means_[0])
            pert = np.sqrt(tmpcv)/2.0
            tmpd = pert * np.ones(d)
        elif ingmm.covariance_type == 'diag':
            foo2 = max(tmpcv)
            pert = np.sqrt(foo2)/2.0
            tmpd = np.zeros(len(tmpmu))
            tmpd[tmpcv==foo2] = pert
        elif ingmm.covariance_type == 'full':
            tD,tV = np.linalg.eig(tmpcv)
            tfoo = max(tD)
            indt = tD.tolist().index(tfoo)
            tmpd = tV[:,indt]*(np.sqrt(tfoo)/2.0)
        else:
            raise ValueError('Not Supported.')

        gmm.means_[ind1] = tmpmu - tmpd
        if ind2 in range(origk):
            gmm.means_[ind2] = tmpmu + tmpd
        else:
            means = gmm.means_.tolist().append(tmpmu+tmpd)
            gmm.means_ = np.asarray(means_)

        myeigs[ind1] = -1

    return gmm


def E_step(ingmm, gmm, opt):
    Nvs = opt['Nvs']
    ncomp_base,dim = np.shape(ingmm.means_)
    
    muLL, muLLcomp, mupost = gmm_ll(ingmm.means_, gmm)
    muLLcomp = muLLcomp - np.log(gmm.weights_)

    muLL1, muLLcomp1, mupost1 = gmm_ll(ingmm.means_, ingmm)
    muLLcomp1 = muLLcomp1 - np.log(ingmm.weights_)

    trLLcomp1 = 4.0*np.ones(np.shape(muLLcomp1))

    trLLcomp = np.zeros(np.shape(muLLcomp))
    for j in range(gmm.n_components):
        if gmm.covariance_type == 'spherical':
            for i in range(ingmm.n_components):
                trLLcomp[i,j] = dim*ingmm.covariances_[i]/gmm.covariances_[j]
        elif gmm.covariance_type == 'diag':
            for i in range(ingmm.n_components):
                tmptr = ingmm.covariances_[i]/gmm.covariances_[j]
                trLLcomp[i,j] = np.sum(tmptr)
        elif gmm.covariance_type == 'full':
            tmpcv = np.linalg.cholesky(gmm.covariances_[j])
            for i in range(ingmm.n_components):
                SigmaInvSigma = solve_chol(tmpcv, ingmm.covariances_[i])
                trLLcomp[i,j] = np.trace(SigmaInvSigma)

    if opt['pimode'] == 'original':
        LLcomp = (Nvs*ingmm.weights_*(muLLcomp - 0.5*trLLcomp).T).T + np.log(gmm.weights_) 
        LLcomp1 =(Nvs*ingmm.weights_*(muLLcomp1- 0.5*trLLcomp1).T).T + np.log(ingmm.weights_)
        LL  = logtrick2(LLcomp)
        LL1 = logtrick2(LLcomp1)

        logY  = LLcomp - LL
        logY1 = LLcomp1 - LL1

        Y  = np.exp(logY)
        Y1 = np.exp(logY1)

        LB = np.sum(Y*(LLcomp - logY))
        entropyLB = np.sum(Y1*(LLcomp1-Y1))

    elif opt['pimode'] == 'preserve':
        LLcomp = Nvs*(muLLcomp - 0.5*trLLcomp) + np.log(gmm.weights_)
        LLcomp1 = Nvs*(muLLcomp1- 0.5*trLLcomp1) + np.log(ingmm.weights_)
        LL  = logtrick2(LLcomp)
        LL1 = logtrick2(LLcomp1)

        logY  = LLcomp - LL
        logY1 = LLcomp1 - LL1

        Y  = np.exp(logY)
        Y1 = np.exp(logY1)

        tmplb = (ingmm.weights_*Y.T).T*(LLcomp - logY)
        LB = np.sum(tmplb)
        entropyLB = np.sum(ingmm.weights_*LLcomp1.T)
    else:
        raise ValueError('Not Supported.')

    return LL,LLcomp,Y,LB,entropyLB


def M_step(gmm, ingmm, Y, opt):
    ''' Update parameters of gmm
    gmm = M_step(gmm, ingmm, Y, opt)'''

    ncomp_base, dim = np.shape(ingmm.means_)
    pi_new = ingmm.weights_.reshape((ncomp_base,1))

    if opt['pimode'] == 'original':
        gmm.weights_ = np.sum(Y,axis=0)/ncomp_base
    elif opt['pimode'] == 'preserve':
        tmpz = pi_new*Y
        gmm.weights_ = np.sum(tmpz,axis=0)/np.sum(tmpz)
    else:
        raise ValueError('Not Supported.')

    W = Y*pi_new
    W = W/np.sum(W,axis=0)

    mumtx = ingmm.means_
    means = []
    covariances = []
    for j in range(gmm.n_components):
        if np.isinf(W[:,j]).any() or np.isnan(W[:,j]).any():
            mu = np.asarray([])
            cv = np.asarray([])
        else:
            Wj = W[:,j]
            mu = np.zeros(dim)
            for c in range(ncomp_base):
                mu = mu + Wj[c]*ingmm.means_[c]

            Xmu = mumtx - mu

            if gmm.covariance_type == 'spherical':
                cvm = np.zeros(dim)
                for c in range(ncomp_base):
                    cvm = cvm + Wj[c]*(ingmm.covariances_[c] + Xmu[c,:]*Xmu[c,:])
                cv = np.sum(cvm)/fload(dim)
                cv = max(cv, opt['cvminreg'])

            elif gmm.covariance_type == 'diag':
                cv = np.zeros(dim)
                for c in range(ncomp_base):
                    cv = cv + Wj[c]*(ingmm.covariances_[c] + Xmu[c,:]*Xmu[c,:])
                cv[cv<opt['cvminreg']] = opt['cvminreg']

            elif gmm.covariance_type == 'full':
                cvtmp = np.zeros_like(ingmm.covariances_[0])
                for jj in range(len(Wj)):
                    cvtmp += Wj[jj] * ingmm.covariances_[jj]

                cv = np.dot((Wj*Xmu.T),Xmu) + cvtmp

                if opt['cvminreg'] > 0:
                    D, V = np.linalg.eig(cv)
                    D[D<opt['cvminreg']] = opt['cvminreg']
                    cv = np.dot(np.dot(V, np.diag(D)), V.T)
            else:
                raise ValueError('Not Supported.')

        means.append(mu)
        covariances.append(cv)

    gmm.means_ = np.asarray(means)
    gmm.covariances_ = np.asarray(covariances)


    return gmm






def logtrick2(LLcomp):
    '''LLcomp: N x K, each column is the log-likelihoods of ingmm.mu on a component 
       of gmm = log(a), log(b) ... log(k)
       output a row vector N x 1 = log(sum(a+b+...k)) 
    '''
    N = np.shape(LLcomp)[0]
    mv = np.max(LLcomp, axis=1)
    mv = mv.reshape((N,1))
    cterm = np.sum(np.exp(LLcomp-mv), axis=1)
    s = mv + np.log(cterm.reshape((N,1)))
    return s





def draw_ellipse(position, covariance, ax=None, **kwargs):
    """Draw an ellipse with a given position and covariance"""
    ax = ax or plt.gca()
    
    # Convert covariance to principal axes
    if covariance.shape == (2, 2):
        U, s, Vt = np.linalg.svd(covariance)
        angle = np.degrees(np.arctan2(U[1, 0], U[0, 0]))
        width, height = 2 * np.sqrt(s)
    else:
        angle = 0
        width, height = 2 * np.sqrt(covariance)
    
    # Draw the Ellipse
    for nsig in range(1, 4):
        ax.add_patch(Ellipse(position, nsig * width, nsig * height,
                             angle, **kwargs))
        
def plot_gmm(gmm, X, label=False, ax=None):
    ax = ax or plt.gca()
    labels = gmm.fit(X).predict(X)
    if label:
        ax.scatter(X[:, 0], X[:, 1], c=labels, s=40, cmap='rainbow', zorder=2)
    else:
        ax.scatter(X[:, 0], X[:, 1], s=40, zorder=2)
    ax.axis('equal')
    
    w_factor = 0.2 / gmm.weights_.max()
    for pos, covar, w in zip(gmm.means_, gmm.covariances_, gmm.weights_):
        draw_ellipse(pos, covar, alpha=w * w_factor)



def gmm_plot1d(gmm, color, opt='cpg', dim=0):
    """plot a gmm in 1d
    INPUTS:
      opt = 'c' -- plot component
          = 'p' -- plot priors
          = 'g' -- plot gmm
      dim = dimension to use [default = 0]"""

    # # select dimension (if necessary)
    # if len(gmm.means_[0]) != 1:
    #     for t in range(len(gmm.means_)):
    #         gmm.means_[t] = gmm.means_[t][dim]
    #         if gmm.covariance_type == 'diag':
    #             gmm.covariances_[t] = gmm.covariances_[t][dim]
    #         elif gmm.covariance_type == 'full':
    #             gmm.covariances_[t] = gmm.covariances_[t][dim, dim]

    stdext = 3
    xlo = []
    xhi = []
    for j in range(gmm.n_components):
        xlo.append(gmm.means_[j][0] - np.sqrt(gmm.covariances_[j][0][0]) * stdext)
        xhi.append(gmm.means_[j][0] + np.sqrt(gmm.covariances_[j][0][0]) * stdext)

    xlo = min(xlo)
    xhi = max(xhi)

    x = np.linspace(xlo, xhi, 200)
    x = x.reshape(-1,1)

    LL, LLcomp, post = gmm_ll(x, gmm)
    # print LL
    for j in range(gmm.n_components):
        if 'c' in opt:
            plt.plot(x, np.exp(LLcomp[:, j]), color+':')
        # if 'p' in opt:
            # text display the priors
        if 'g' in opt:
            plt.plot(x, np.exp(LL), color+'-')



def gmm_ll(X, gmm):
    """ gmm_ll -- log-likelihood of GMM

        USAGE: 
        LL, LLcomp, post = gmm_ll(X, gmm)

        INPUTS:
          X -- 2d array(matrix) with each row the mean of a GMM component
          gmm -- GMM model
        
        OUTPUTS:
          LL     -- log-likelihood of X        [n x 1]
                    = log(gmm(X))
          LLcomp -- component log-likelihoods  [n x K]
          post   -- posterior probabilities    [n x K]

        if bkgndclass is used, LLcomp and post are [n x (K+1)], where the 
        last column is the log-likelihood and posterior in the background class"""

    K = gmm.n_components
    N,d = np.shape(X)

    LLcomp = np.zeros((N, K))
    # print 'gmm = {0}\n'.format(gmm)
    for c in range(K):
        tmpLL = logComponent(gmm, X, c, d, N)
        LLcomp[:, c] = tmpLL

    LL = logtrick2(LLcomp)
    # LL should be a column vector
    # return LL, LLcomp

    post = np.exp(LLcomp - LL)
    return LL, LLcomp, post


def logComponent(gmm, X, c, d, N):
    # return the log-likelihood of component c
    # myLLcomp = logComponent(gmm, X, c, d, N)
    # myLLcomp should be a row vector

    mu = gmm.means_[c]
    # print 'mu = {0}\n'.format(mu)
    cv = gmm.covariances_[c]

    tempx = X - mu

    if gmm.covariance_type == 'spherical':
        g = np.sum(tempx * tempx)/cv
        ld = d * np.log(cv)

    elif gmm.covariance_type == 'diag':
        # if cv is diagonal matrix, it's stored as a 1d array in a row
        tmp = tempx/np.sqrt(cv)
        g = np.sum(tmp * tmp, axis=1)
        ld = np.sum(np.log(cv))

    elif gmm.covariance_type == 'full':
        L = np.linalg.cholesky(cv)
        InvCVX = solve_chol(L, tempx.T)
        g = np.sum(InvCVX.T * tempx, axis=1)
        # ld = np.linalg.slogdet(cv)[1]  # this function returns a tuple with two elements: sign and value
        ld = logdet_chol(L)
    else:
        raise ValueError('Not Supported.')

    # print 'gmm.K = {0},  c = {1}, gmm.pi = {2}'.format(gmm['K'], c, gmm['pi'])
    # if gmm['pi'][c]:

    myLLcomp = -0.5*g - (d/2.0)*np.log(2.0*np.pi) - 0.5*ld + np.log(gmm.weights_[c])

    # else:
    #     myLLcomp = -0.5*g - (d/2.0)*np.log(2.0*np.pi) - 0.5*ld
    return myLLcomp


def logdet_chol(L):
    """log(det(A)) where A = L*L^T with L lower triangular"""
    return 2.0*np.sum(np.log(np.diag(L)))

  

from scipy.linalg.lapack import get_lapack_funcs
# potrs, = get_lapack_funcs(('potrs',),(np.array([0.0]),))


def solve_chol(L, b):
    """inv(A)*b where A = L*L^T with L lower triangular"""
    potrs, = get_lapack_funcs(('potrs',),(L,))
    b, info = potrs(L, b, lower=True, overwrite_b=False)
    if info < 0:
        msg = "Argument %d to lapack's ?potrs() has an illegal value." % info
        raise TypeError(msg)
    if info > 0:
        msg = "Unknown error occured int ?potrs(): error code = %d" % info
        raise TypeError(msg)
    return b



def inv_chol(L):
    """inv(A) where A = L*L^T with L lower triangular"""
    invL = np.linalg.inv(L)
    return np.dot(invL.T, invL)


def logsumexp(x, axis=0):
    np.seterr(all='ignore')
    xShape = x.shape
    if len(xShape) == 0 or xShape[axis] == 0:
        return -np.inf
    alpha = np.amax(x, axis) + (2.0*np.log(x.shape[axis]) - __halflogmax)
    badx = np.logical_not(np.isfinite(x))
    xx = np.exp(x - alpha.reshape(xShape[0:axis] + (1,) + xShape[(axis+1):]))
    xx[badx] = 0
    sumexpx = np.sum(xx, axis)
    if np.isscalar(sumexpx):
        if sumexpx > 0:
            return alpha + np.log(sumexpx)
        else:
            return -np.inf
    else:
        ret = np.empty_like(sumexpx)
        ret[sumexpx > 0] = np.log(sumexpx[sumexpx>0])
        ret[sumexpx <= 0] = -np.inf 
        return alpha + ret



def my_weighted_kmeans(point, weights, init_cluster_center, ncluster, it_max):
    """ my_weighted_kmeans - perform weighted kmeans clustering on "data" with 
        respect to "weights" assigned on each data point.

        INPUT:
            data = d x n matrix
            weights = d x 1 vector
            init_cluster_center = d x ncluster matrix
            ncluster = number of clusters required
            it_max = maximum iteration

        OUTPUT:
            cluster_label = n x 1 vector, indicating the cluster of each data point
            cluster_center = d x ncluster matrix
    """

    it_num = 0
    cluster_num = ncluster
    cluster_center = init_cluster_center
    dim_num, point_num = np.shape(point)

    dist2cluster = np.zeros((cluster_num,point_num))

    # assign each data point to the nearest cluster
    for j in np.arange(cluster_num):
        curr_cluster_center = cluster_center[:,j].reshape(dim_num,1)
        dist2cluster[j,:] = np.sum((point-curr_cluster_center)**2,0)

    cluster = np.argmin(dist2cluster,0)


    # determine the cluster population and weights
    # compute the weighted cluster center
    cluster_center,cluster_population,cluster_weight = gcentroids(point,cluster,weights,cluster_num)


    # set the point energies
    #   - 1. compute the distance of each point to its assigned cluster center
    #   - 2. adjust the point energies by a weight factor
    f,cluster_energy = genergy(point,weights,cluster_center,cluster_weight,cluster)


    old_energy = np.sum(cluster_energy)
    energy_in_iteration = []
    energy_in_iteration.append(old_energy)


    while it_num < it_max:
    # for each cluster
    # find non members ~m
    # compute f(~m) to current cluster -> f'(~m)
    # save these energy f in a k x n matrix, with each element storing the energy of each point to corresponding cluster
    # find minimum energy for each point
    # assign all point to each minimum energy cluster
    # update cluster center, weight, population and energy
    #     print '\niteration = {0}\n'.format(it_num)

        fmat = np.zeros((cluster_num,point_num))
        for j in np.arange(cluster_num):
            members =  [i for i,x in enumerate(cluster) if x==j]
            nonmembers = [i for i,x in enumerate(cluster) if x!=j]
            fmat[j,members] = f[members].reshape(-1,)
            adjust_weight = cluster_weight[j]/(cluster_weight[j]+weights[nonmembers])
            f_nonmembers = np.sum((point[:,nonmembers]-cluster_center[:,j].reshape(-1,1))**2,0)
            fmat[j,nonmembers] = f_nonmembers

        cluster = np.argmin(fmat,0)

        cluster_center,cluster_population,cluster_weight = gcentroids(point, cluster, weights, cluster_num)
        f,cluster_energy = genergy(point,weights,cluster_center,cluster_weight,cluster)
        new_energy = np.sum(cluster_energy)
        if np.abs(new_energy-old_energy) < 1e-6:
            break
        else:
            old_energy = new_energy
            it_num = it_num + 1
            energy_in_iteration.append(new_energy)
        
    return cluster, cluster_center



def gcentroids(point, cluster, weight, cluster_num):
    ndim,npoint = np.shape(point)
    cluster_population = np.zeros((cluster_num,1))
    cluster_weight = np.zeros((cluster_num,1))
    centroids = np.zeros((ndim,cluster_num))

    for j in np.arange(cluster_num):
        members = [i for i,x in enumerate(cluster) if x==j]
        cluster_population[j] = len(members)
        cluster_weight[j]     = np.sum(weight[members])

        centroids[:,j]        = np.sum(point[:,members]*weight[members],1)

        if cluster_weight[j] > 0:
            centroids[:,j] = centroids[:,j]/cluster_weight[j] # cluster center
      

    return centroids, cluster_population, cluster_weight



def genergy(point,weight,cluster_center,cluster_weight,cluster):

    ndim,ndata = np.shape(point)
    cluster_num  = np.shape(cluster_center)[1]

    f= np.zeros((ndata,1))
    cluster_energy = np.zeros((cluster_num,1))

    for j in np.arange(cluster_num):
        members = [i for i,x in enumerate(cluster) if x==j]
        f[members] = np.sum((point[:,members]-cluster_center[:,j].reshape(ndim,1))**2,0).reshape(-1,1)
        cluster_energy[j] = np.sum(weight[members]*f[members])
        f[members] = f[members]*cluster_weight[j]/(cluster_weight[j]-weight[members]).reshape(-1,1)
    
    return f,cluster_energy





# coding: utf-8
"""

    CustomMessageBox.py

    Copyright (c) 2018-2020, Masatsuyo Takahashi, KEK-PF

    https://msdn.microsoft.com/ja-jp/library/windows/desktop/ms645505(v=vs.85).aspx
    https://stackoverflow.com/questions/27257018/python-messagebox-with-icons-using-ctypes-and-windll
    https://www.skyarch.net/blog/?p=4812
    http://www.jasinskionline.com/windowsapi/ref/b/bm_click.html
    https://stackoverflow.com/questions/20995045/how-to-include-image-in-message-box-using-ctypes-in-python/21051586#21051586
"""
import os
import ctypes

SetActiveWindow     = ctypes.windll.user32.SetActiveWindow
SendMessage         = ctypes.windll.user32.SendMessageW
PostMessage         = ctypes.windll.user32.PostMessageW
MessageBox          = ctypes.windll.user32.MessageBoxW
FindWindow          = ctypes.windll.user32.FindWindowW
FindWindowEx        = ctypes.windll.user32.FindWindowExW
GetWindowRect       = ctypes.windll.user32.GetWindowRect
SetWindowPos        = ctypes.windll.user32.SetWindowPos

# icons
MB_OK               = 0x0
MB_OKCANCEL         = 0x01
MB_YESNOCANCEL      = 0x03
MB_YESNO            = 0x04
MB_HELP             = 0x4000
ICON_EXLAIM         = 0x30
ICON_INFORMATION    = 0x40
ICON_ERROR          = 0x10
ICON_WARNING        = 0x30
ICON_QUESTION       = 0x20
IDOK    = 1
IDCANCEL = 2
IDYES   = 6

from HookableMessageBox     import MyMessageBox, AlterIcon, AlterButton
# from TestUtils              import _RECT

class _RECT( ctypes.Structure ):
    _fields_ =  [
                    ( 'left',    ctypes.c_int ),
                    ( 'top',     ctypes.c_int ),
                    ( 'right',   ctypes.c_int ),
                    ( 'bottom',  ctypes.c_int ),
                ]

class InvokableButton:
    def __init__(self, parent, title):
        self.parent = parent
        self.title  = title

    def invoke(self):
        _hWndP  = FindWindow( 0, self.title )
        _hWndB  = FindWindowEx(_hWndP, 0, 0, "OK")
        SetActiveWindow(_hWndP)
        SendMessage(_hWndB, 0x00F5, 0, 0)       # 0x00F5 - BM_CLICK
        # PostMessage(_hWndB, 0x00F5, 0, 0)

current_messagebox = None

def get_current_messagebox():
    return current_messagebox

class CustomMessageBox_:
    def __init__(self, **options):
        self.options    = options
        self.parent     = options.get( 'parent' )
        self.title      = options.get( 'title' )
        self.message    = options.get( 'message' )
        self.type_      = options.get( 'type_' )
        self.icon_path  = options.get( 'icon_path' )
        self._hWnd      = 0 if self.parent is None else self.parent.winfo_id()
        self.ok_button  = InvokableButton( self, self.title )
        rect    = _RECT( 0, 0, 0, 0 )
        ret     = GetWindowRect( self._hWnd, ctypes.byref(rect) )
        # print( '__init__', rect.left, rect.top, rect.right, rect.bottom )
        self.parent_x   = ( rect.left + rect.right ) // 2
        self.parent_y   = ( rect.top + rect.bottom ) // 2

    def show( self ):
        icon_path = self.icon_path
        if icon_path is None:
            icon_path = os.environ.get('OUR_ICON_PATH')

        self.counter = 0

        def hook_callback( _hWnd ):
            if self.counter > 0:
                return

            self.counter += 1
            if icon_path is not None:
                AlterIcon( _hWnd, icon_path )
            # AlterButton( _hWnd )
            rect    = _RECT( 0, 0, 0, 0 )
            ret     = GetWindowRect( _hWnd, ctypes.byref(rect) )
            # print( 'hook_callback', rect.left, rect.top, rect.right, rect.bottom )
            width   = rect.right - rect.left
            height  = rect.bottom - rect.top
            ret = SetWindowPos( _hWnd, 0, self.parent_x, self.parent_y, width, height, 0 )
            # print( 'ret=', ret )
            if ret == 0:
                print( ctypes.FormatError() )

        ret = MyMessageBox(self._hWnd, self.message, self.title, self.type_, hook_callback)
        return ret

"""
    Only a part of functions are implemented. Remember to add ones if required.
"""

def showinfo( title, message, **options ):
    global current_messagebox
    type_ = MB_OK | ICON_INFORMATION
    current_messagebox = CustomMessageBox_( title=title, message=message, type_=type_, **options )
    ret = current_messagebox.show()
    return ret == IDOK

def showerror( title, message, **options ):
    global current_messagebox
    type_ = MB_OK | ICON_ERROR
    current_messagebox = CustomMessageBox_( title=title, message=message, type_=type_, **options )
    current_messagebox.show()

def showwarning( title, message, **options ):
    global current_messagebox
    type_ = MB_OK | ICON_WARNING
    current_messagebox = CustomMessageBox_( title=title, message=message, type_=type_, **options )
    current_messagebox.show()

def askyesno(title=None, message=None, **options):
    global current_messagebox
    type_ = MB_YESNO | ICON_QUESTION
    current_messagebox = CustomMessageBox_( title=title, message=message, type_=type_, **options )
    ret = current_messagebox.show()
    return ret == IDYES

def askokcancel(title=None, message=None, **options):
    global current_messagebox
    type_ = MB_OKCANCEL | ICON_QUESTION
    current_messagebox = CustomMessageBox_( title=title, message=message, type_=type_, **options )
    ret = current_messagebox.show()
    return ret == IDOK



# coding: utf-8
"""

    TimeUtils.py

    Copyright (c) 2019-2022, Masatsuyo Takahashi, KEK-PF

"""
from datetime import datetime, timedelta
from time import sleep

"""
    from https://stackoverflow.com/questions/538666/format-timedelta-to-string
"""
def format_seconds(seconds):
    hours, remainder = divmod(seconds, 3600)
    minutes, seconds = divmod(remainder, 60)
    return '%d:%0d:%02d' % (hours, minutes, seconds)

def seconds_to_datetime(seconds):
    return datetime.strptime(format_seconds(seconds), "%H:%M:%S")

def friendly_time_str(given_time):
    format  = "%Y-%m-%d %H:%M"
    currn_str = datetime.now().strftime(format)
    given_str = given_time.strftime(format)

    hhmm = given_str[11:]
    if given_str[0:10] == currn_str[0:10]:
        ret_str = hhmm
    else:
        if given_str[0:6] == currn_str[0:6]:
            dd1 = int(given_str[8:10])
            dd0 = int(currn_str[8:10])
            n = dd1 - dd0
            if n == 1:
                ret_str = "tomorrow " + hhmm
            else:
                mm1 = int(given_str[5:7])
                ret_str = "%d/%d %s" % (mm1, dd1, hhmm)
        else:
            ret_str = given_str
    return ret_str

def wait_until_hhmm(yyyy_mm_dd_hh_mm, interval=60, debug=False):
    while True:
        datetime_str = datetime.now().strftime("%Y-%m-%d %H:%M")
        if debug:
            print(datetime_str)
        if datetime_str >= yyyy_mm_dd_hh_mm:
            break
        sleep(interval)

if __name__ == '__main__':
    datetime_str = (datetime.now() + timedelta(minutes=1)).strftime("%Y-%m-%d %H:%M")
    wait_until_hhmm(datetime_str, debug=True)



"""
    Theory.SpherePlot.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.integrate import quad
from scipy.special import spherical_jn


def sphere_xyz(center=(0,0,0), r=1, n=20, u=None, v=None):
    """
    https://matplotlib.org/mpl_toolkits/mplot3d/tutorial.html
    """
    if u is None:
        u = np.linspace(0, 2 * np.pi, n)
    if v is None:
        v = np.linspace(0, np.pi, n)
    x = r * np.outer(np.cos(u), np.sin(v))
    y = r * np.outer(np.sin(u), np.sin(v))
    z = r * np.outer(np.ones(np.size(u)), np.cos(v))
    return center[0]+x, center[1]+y, center[2]+z


def demo1(seed=1111):
    import molass_legacy.KekLib.DebugPlot as plt

    fig = plt.figure(figsize=(10,10))
    ax = fig.add_subplot(111, projection='3d')

    R = 0.3
    v0 = 4*np.pi*R**3/3
    ax.set_title(r"$N=16, V=4^3=64, v_1=\frac{V}{N}=4, v_0=\frac{4}{3}\pi R^3=%.3g$" % v0, fontsize=20)

    np.random.seed(seed)
    purtubations = np.random.uniform(-0.5, 0.5, (16,3))
    for k, center in enumerate([
                    (0.5, 0.5, 0.5), (2.5, 0.5, 0.5), (0.5, 0.5, 2.5), (2.5, 0.5, 2.5),
                    (1.5, 1.5, 1.5), (3.5, 1.5, 1.5), (1.5, 1.5, 3.5), (3.5, 1.5, 3.5), 
                    (0.5, 2.5, 0.5), (2.5, 2.5, 0.5), (0.5, 2.5, 2.5), (2.5, 2.5, 2.5),
                    (1.5, 3.5, 1.5), (3.5, 3.5, 1.5), (1.5, 3.5, 3.5), (3.5, 3.5, 3.5), 
                    ]):
        p = purtubations[k]
        s = sphere_xyz(center+p, R)
        ax.plot_surface(*s, color='b')

    x, y, z = np.indices((4, 4, 4))
    cubes = x + y + z  >= 0
    colors = np.array([[['#1f77b430']*4]*4]*4)

    ax.voxels(cubes, facecolors=colors)

    fig.tight_layout()
    plt.show()

if __name__ == '__main__':
    import sys
    import seaborn as sns
    sns.set_theme()
    sys.path.append("../lib")
    demo1()


"""
    Optimizer.FvScoreConverter.py

    Copyright (c) 2023-2024, SAXS Team, KEK-PF
"""
import numpy as np

def sigmoid(x, L ,x0, k, b):
    if not np.isscalar(x):
        # see https://stackoverflow.com/questions/59297543/why-do-i-get-the-loop-of-ufunc-does-not-support-argument-0-of-type-int-error-f
        x = x.astype(float)
    return L/(1 + np.exp(-k*(x-x0))) + b

def convert_score(value):
    return sigmoid(value, -200, 0, 1.5, 100)

convert_scores = np.vectorize(convert_score)

def invert_score(sv):
    from scipy.optimize import root

    assert np.isscalar(sv)

    def fun(x):
        return [convert_score(x)[0] - sv]

    sol = root(fun, (0, ) )
    return sol.x[0]

def compute_fv_adjustment(sv, target_sv):
    return invert_score(target_sv) - invert_score(sv)

if __name__ == '__main__':
    import matplotlib.pyplot as plt
    import seaborn
    seaborn.set()
    x = np.linspace(-5, 5)
    print(x.shape)
    y = convert_score(x)

    y_ = y[[20, 30]]
    x_ = [invert_score(v) for v in y_]

    print("compute_fv_adjustment: ", compute_fv_adjustment(-75, 50))

    fig, ax = plt.subplots()
    ax.set_title("FV Conversion to Scale of Hundred (SV)", fontsize=20)
    ax.set_xlabel("FV", fontsize=16)
    ax.set_ylabel("SV", fontsize=16)
    ax.plot(x, y, label="conversion")
    ax.plot(x_, y_, "o", label="inversion verification")
    ax.legend()
    fig.tight_layout()
    plt.show()
